# -*- coding: utf-8 -*-
"""
Module to build and save least cost raster layers
"""
import logging
import numpy as np
import os

from reV.handlers.exclusions import ExclusionLayers

from reVX.handlers.geotiff import Geotiff
from reVX.utilities.exclusions_converter import ExclusionsConverter
from reVX.least_cost_xmission.config import XmissionConfig

logger = logging.getLogger(__name__)

NLCD_LAND_USE_CLASSES = {
    'cropland': [80, 81],
    'forest': [41, 42, 43],
    'wetland': [90, 95],
    'suburban': [21, 22, 23],
    'urban': [24],
}


class XmissionCostCreator(ExclusionsConverter):
    """
    Class to create and save Transmission cost layers to a .h5 Exclusion file
    - dist_to_coast (for valid sc_points)
    - base_costs
    - multiplier_*mw
    - xmission_barrier
    """
    PIXEL_SIZE = 90

    def __init__(self, h5_fpath, iso_regions_fpath, iso_lookup=None):
        """
        Parameters
        ----------
        h5_fpath : str
            H5 file to save costs to
        iso_regions_fpath : str
            ISO regions raster file
        iso_lookup : dict, optional
            Keys are ISO names used in various dicts with keys being int codes
            use in the iso_regions_fpath, by default None
        """
        super().__init__(h5_fpath)

        logger.debug('Loading ISO regions')
        with Geotiff(iso_regions_fpath) as gt:
            self._iso_regions = gt.values[0].astype('uint16')
            self._profile = gt.profile

        self._profile['dtype'] = 'float32'
        self._profile['nodata'] = None

        self._iso_lookup = iso_lookup if iso_lookup is not None else {}

        if not os.path.exists(self._excl_h5):
            self._init_h5(self._excl_h5, iso_regions_fpath)

        if 'ISO_regions' not in self.layers:
            self._geotiff_to_h5(self._excl_h5, 'ISO_regions',
                                iso_regions_fpath)
        else:
            self._check_geotiff(self._excl_h5, iso_regions_fpath)

    @staticmethod
    def _compute_slope_mult(slope, config=None):
        """
        Create slope multiplier raster for a region. Unspecified slopes are
        left at 1.0

        Parameters
        ----------
        slope : numpy.ndarray
            Slope raster clipped to a region- "Terrain slope in % of grade"
        config : dict | None
            Multipliers and slope cut offs for hilly and mountain terrain.
            Use defaults if None.
            config['hill_mult'] : float
                Multiplier for hilly terrain
            config['mtn_slope'] : float
                Multiplier for mountainous terrain
            config['hill_slope'] : float
                Slope at and above which a cell is classified as hilly.
            config['mtn_slope'] : float
                Slope at and above which a cell is classified as mountainous

        Returns
        -------
        numpy.ndarray
            Slope multiplier raster. Minimum value for any cell is 1.
        """
        if config is None:
            config = {}

        hill_mult = config.get('hill_mult', 1)
        mtn_mult = config.get('mtn_mult', 1)
        hill_slope = config.get('hill_slope', 2)
        mtn_slope = config.get('mtn_slope', 8)

        hilly = (slope >= hill_slope) & (slope < mtn_slope)
        mountainous = slope >= mtn_slope

        mult_raster = np.ones(slope.shape, dtype=np.float32)
        mult_raster[hilly] = hill_mult
        mult_raster[mountainous] = mtn_mult

        return mult_raster

    @staticmethod
    def _compute_land_use_mult(land_use, multipliers, land_use_classes):
        """
        Convert NLCD raster to land use multiplier raster for a single region.
        Land classes without specified multipliers are left a 1.

        Parameters
        ----------
        land_use : numpy.ndarray
           NLCD land user raster clipped to a region
        multipliers : dict
            Multiplier for for land classes, E.g. {'forest': 1.5}
        land_use_classes : dict
            NCLD land use codes corresponding to use classes for multipliers

        Returns
        -------
        numpy.ndarray
            Land use multiplier raster. Minimum value for any cell is 1.
        """
        mult_raster = np.ones(land_use.shape, dtype=np.float32)

        indices = []  # [(index0, multiplier0, _class_value0), ...]
        for _class, multiplier in multipliers.items():
            if _class not in land_use_classes:
                msg = (f'Class {_class} not in land_use_classes: '
                       f'{land_use_classes}')
                raise ValueError(msg)

            values = land_use_classes[_class]
            assert isinstance(values, list)

            for value in values:
                index = land_use == value
                indices.append((index, multiplier, value))

        for i in indices:
            mult_raster[i[0]] = i[1]

        return mult_raster

    def compute_multipliers(self, iso_mults, excl_h5=None,
                            slope_layer='srtm_slope',
                            nlcd_layer='usa_mrlc_nlcd2011',
                            land_use_classes=None,
                            default_mults=None):
        """
        Create costs multiplier raster

        Parameters
        ----------
        iso_mults : dict
            Land use and slope multipliers for ISOs
        excl_h5 : str, optional
            Path to exclusion .h5 file containing NLCD and slope layers,
            if None assume NLCD and slope layers are in self._excl_h5,
            by default None
        sloper_layer : str, optional
            Name of slope layer in excl_h5, by default 'srtm_slope'
        ncld_layer : str, optional
            Name of NLCD (land use) layer in excl_h5, by default
            'usa_mrlc_nlcd2011'
        land_use_classes : dict, optional
            NCLD land use codes corresponding to use classes for multipliers.
            if None us NLCD_LAND_USE_CLASSES. by default None
        default_mults : dict, optional
            Multipliers for regions not specified in iso_mults_fpath.
            by default None

        Returns
        -------
        numpy.ndarray
            Costs multiplier raster, including slope and land use
        """
        if excl_h5 is None:
            excl_h5 = self._excl_h5

        logger.debug('Loading slope and land use rasters')
        with ExclusionLayers(excl_h5) as el:
            slope = el[slope_layer]
            land_use = el[nlcd_layer]

        logger.debug('Loading complete')

        assert self._iso_regions.shape == land_use.shape == slope.shape, \
            'All rasters must be the same shape'

        mults_arr = np.ones(self._iso_regions.shape, dtype=np.float32)
        regions_mask = np.full(mults_arr.shape, False, dtype=bool)

        if land_use_classes is None:
            land_use_classes = NLCD_LAND_USE_CLASSES

        for r_conf in iso_mults:
            iso = r_conf['iso']
            logger.info(f'Processing multipliers for region {iso}')

            if self._iso_lookup is not None:
                iso_name = iso
                iso = self._iso_lookup[iso]
                logger.debug(f'{iso_name} has id {iso}')

            mask = self._iso_regions == iso
            regions_mask = regions_mask | mask

            if 'land_use' in r_conf:
                r_lu = land_use[mask]
                lu_mult = self._compute_land_use_mult(r_lu, r_conf['land_use'],
                                                      land_use_classes)
                mults_arr[mask] = lu_mult

            if 'slope' in r_conf:
                r_slope = slope[mask]
                slope_mult = self._compute_slope_mult(r_slope, r_conf['slope'])
                mults_arr[mask] = mults_arr[mask] * slope_mult

        # Calculate multipliers for regions not defined in `config`
        logger.debug('Processing default region')

        if default_mults is None:
            default_mults = {}

        default_mask = ~regions_mask

        if 'land_use' in default_mults:
            rlu = land_use[default_mask]
            lu_mult = self._compute_land_use_mult(rlu,
                                                  default_mults['land_use'],
                                                  land_use_classes)
            mults_arr[default_mask] = lu_mult

        if 'slope' in default_mults:
            r_slope = slope[default_mask]
            slope_mult = self._compute_slope_mult(r_slope,
                                                  default_mults['slope'])
            mults_arr[default_mask] = mults_arr[default_mask] * slope_mult

        # Set water multiplier last so we don't get super high multipliers at
        # water body boundaries next to steep slopes
        mults_arr[land_use == 11] = 10

        return mults_arr

    def compute_base_line_costs(self, capacity, base_line_costs):
        """
        Get base line cost per cell raster for a given voltage

        Parameters
        ----------
        capacity : int
            Desired line capacity (MW). This must be in base_line_costs.
        base_line_costs : dict
            Base line cost per mile for ISOs

        Returns
        -------
        base_cost : numpy.ndarray
            Cost per cell raster in same shape as iso_regions
        """
        base_cost = -1 * np.ones(self._iso_regions.shape, dtype=np.float32)

        for iso in base_line_costs:
            logger.info(f'Processing costs for {iso} for {capacity}MW')
            iso_code = self._iso_lookup[iso]
            cost_per_mile = base_line_costs[iso][str(capacity)]
            cost_per_cell = cost_per_mile / 1609.344 * self.PIXEL_SIZE
            logger.debug(f'$/mile is {cost_per_mile}, $/cell is '
                         f'{cost_per_cell}')
            mask = self._iso_regions == iso_code
            base_cost[mask] = cost_per_cell

        return base_cost

    def save_layer(self, layer_name, data):
        """
        Save data to h5 file

        Parameters
        ----------
        layer_name : str
            Name of layer to save in h5
        data : np.ndarray
            Data to savein h5
        """
        logger.debug(f'Saving {layer_name} to h5')
        self._profile['nodata'] = None
        self._write_layer(self._excl_h5, layer_name, self._profile, data)

    def create_geotiff(self, geotiff_fpath, data):
        """
        Save geotiff to disk

        Parameters
        ----------
        tiff_dir : str
            Directory to save tiffs in
        geotiff_fpath : str
            File name of tiff including extension
        data : np.ndarray
            Data to save
        """
        nodata = self._profile['nodata']
        logger.debug(f'Saving {geotiff_fpath}')
        self._write_geotiff(geotiff_fpath, self._profile, data)

        # _write_geotiff sets nodata to the max for the dtype. For float32 this
        # then breaks JSONifying the profile when writing h5s. Reset it.
        self._profile['nodata'] = nodata

    @classmethod
    def run(cls, h5_fpath, iso_regions_fpath, excl_h5=None,
            cost_configs=None, slope_layer='srtm_slope',
            nlcd_layer='usa_mrlc_nlcd2011', default_mults=None, tiff_dir=None,
            extra_layers=None):
        """
        Build cost rasters using base line costs and multipliers. Save to
        h5 file and tiff if desired

        Parameters
        ----------
        h5_fpath : str
            H5 file to save costs to
        iso_regions_fpath : str
            File with raster of ISO regions
        excl_h5 : str, optional
            Path to exclusion .h5 file containing NLCD and slope layers,
            if None assume NLCD and slope layers are in self._excl_h5,
            by default None
        cost_configs : str | dict, optional
            Path to json file containing Xmission cost configuration values,
            or jsonified dict of cost configuration values,
            or dictionary of configuration values,
            or dictionary of paths to config jsons,
            if None use defaults, by default None
        sloper_layer : str, optional
            Name of slope layer in excl_h5, by default 'srtm_slope'
        ncld_layer : str, optional
            Name of NLCD (land use) layer in excl_h5, by default
            'usa_mrlc_nlcd2011'
        default_mults : dict, optional
            Multipliers for regions not specified in iso_mults_fpath.
            by default None
        tiff_dir : str | None
            Path to save costs and intermediary rasters as geotiffs in. Don't
            save to geotiff if None.
        extra_layers : dict, optional
            Extra layers to add to h5 file, for example dist_to_coast. Format
            of dictionary is:
            {'layers': ['layers1', ...]
             'descriptions': optional,
             'scale_factors': optional,
             'kwargs': optional}
            by default None.
        """
        xc = XmissionConfig(config=cost_configs)

        xcc = cls(h5_fpath, iso_regions_fpath, iso_lookup=xc['iso_lookup'])

        if tiff_dir is None:
            save_geotiff = False
        else:
            save_geotiff = True

        mults_arr = xcc.compute_multipliers(
            xc['iso_multipliers'],
            excl_h5=excl_h5,
            slope_layer=slope_layer,
            nlcd_layer=nlcd_layer,
            land_use_classes=xc['land_use_classes'],
            default_mults=default_mults)

        xcc.save_layer('tie_line_multipliers', mults_arr)

        if save_geotiff:
            tiff_path = os.path.join(tiff_dir, 'multipliers.tif')
            xcc.create_geotiff(tiff_path, mults_arr)

        for power_class, capacity in xc['power_classes'].items():
            logger.info(f'Calculating costs for class {power_class} using a '
                        f'{capacity}MW line')
            blc_arr = xcc.compute_base_line_costs(capacity,
                                                  xc['base_line_costs'])

            if save_geotiff:
                tiff_path = os.path.join(tiff_dir,
                                         f'base_line_costs_{capacity}MW.tif')
                xcc.create_geotiff(tiff_path, blc_arr)

            costs_arr = blc_arr * mults_arr

            xcc.save_layer(f'tie_line_costs_{capacity}MW', costs_arr)

            if save_geotiff:
                tiff_path = os.path.join(tiff_dir,
                                         f'tie_line_costs{capacity}MW.tif')
                xcc.create_geotiff(tiff_path, costs_arr)

        if extra_layers:
            layers = extra_layers['layers']
            descriptions = extra_layers.get('descriptions', None)
            scale_factors = extra_layers.get('scale_factors', None)
            kwargs = extra_layers.get('kwargs', {})
            ExclusionsConverter.layers_to_h5(h5_fpath, layers,
                                             descriptions=descriptions,
                                             scale_factors=scale_factors,
                                             **kwargs)



# -*- coding: utf-8 -*-
"""
Compute setbacks exclusions
"""
from abc import ABC, abstractmethod
from concurrent.futures import as_completed
from warnings import warn
import os
import logging
import pathlib
import numpy as np
import geopandas as gpd
from rasterio import features
from shapely.geometry import shape

from rex.utilities import parse_table, SpawnProcessPool, log_mem
from reV.handlers.exclusions import ExclusionLayers
from reVX.utilities.exclusions_converter import ExclusionsConverter
from reVX.utilities.utilities import log_versions

logger = logging.getLogger(__name__)


class BaseSetbacks(ABC):
    """
    Create exclusions layers for setbacks
    """

    def __init__(self, excl_fpath, base_setback_dist, regulations_fpath=None,
                 multiplier=None, hsds=False, chunks=(128, 128)):
        """
        Parameters
        ----------
        excl_fpath : str
            Path to .h5 file containing exclusion layers, will also be
            the location of any new setback layers
        base_setback_dist : float | int
            Base setback distance (m). This value will be used to
            calculate the setback distance when a multiplier is provided
            either via the `regulations_fpath`csv or the `multiplier`
            input. In this case, the setbacks will be calculated using
            `base_setback_dist * multiplier`.
        regulations_fpath : str | None, optional
            Path to regulations .csv file. At a minimum, this csv must
            contain the following columns: `Value Type`, which
            specifies wether the value is a multiplier or static height,
            `Value`, which specifies the numeric value of the setback or
            multiplier, and `FIPS`, which specifies a unique 5-digit
            code for each county (this can be an integer - no leading
            zeros required). Typically, this csv will also have a
            `Feature Type` column that labels the type of setback
            that each row represents. Valid options for the `Value Type`
            are:
                - "Max-tip Height Multiplier"
                - "Rotor-Diameter Multiplier"
                - "Hub-height Multiplier"
                - "Structure Height multiplier"
                - "Meters"
            If this input is `None`, a generic setback of
            `base_setback_dist * multiplier` is used. By default `None`.
        multiplier : int | float | None, optional
            A setback multiplier to use if regulations are not supplied.
            This multiplier will be applied to the plant height. If
            supplied along with `regulations_fpath`, this input will be
            ignored. By default `None`.
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on
            AWS behind HSDS. By default `False`.
        chunks : tuple, optional
            Chunk size to use for setback layers, if None use default
            chunk size in excl_fpath. By default `(128, 128)`.
        """
        log_versions(logger)
        self._base_setback_dist = base_setback_dist
        self._excl_fpath = excl_fpath
        self._hsds = hsds
        excl_props = self._parse_excl_properties(excl_fpath, chunks, hsds=hsds)
        self._shape, self._chunks, self._profile = excl_props

        self._regulations, self._multi = self._preflight_check(
            regulations_fpath, multiplier
        )

    def __repr__(self):
        msg = "{} for {}".format(self.__class__.__name__, self._excl_fpath)
        return msg

    @staticmethod
    def _parse_excl_properties(excl_fpath, chunks, hsds=False):
        """Parse shape, chunk size, and profile from exclusions file.

        Parameters
        ----------
        excl_fpath : str
            Path to .h5 file containing exclusion layers, will also be
            the location of any new setback layers
        chunks : tuple | None
            Chunk size of exclusions datasets
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on
            AWS behind HSDS. By default `False`.

        Returns
        -------
        shape : tuple
            Shape of exclusions datasets
        chunks : tuple | None
            Chunk size of exclusions datasets
        profile : str
            GeoTiff profile for exclusions datasets
        """
        with ExclusionLayers(excl_fpath, hsds=hsds) as exc:
            dset_shape = exc.shape
            profile = exc.profile
            if chunks is None:
                chunks = exc.chunks

        if len(chunks) < 3:
            chunks = (1, ) + chunks

        if len(dset_shape) < 3:
            dset_shape = (1, ) + dset_shape

        logger.debug('Exclusions properties:\n'
                     'shape : {}\n'
                     'chunks : {}\n'
                     'profile : {}\n'
                     .format(dset_shape, chunks, profile))

        return dset_shape, chunks, profile

    def _preflight_check(self, regulations_fpath, multiplier):
        """Apply preflight checks to the regulations path and multiplier.

        Run preflight checks on setback inputs:
        1) Ensure either a regulations .csv is provided, or
           a setback multiplier
        2) Ensure regulations has county FIPS, map regulations to county
           geometries from exclusions .h5 file

        Parameters
        ----------
        regulations_fpath : str | None
            Path to regulations .csv file, if `None`, create global
            setbacks.
        multiplier : int | float | str | None
            Setback multiplier to use if regulations are not supplied.

        Returns
        -------
        regulations: `geopandas.GeoDataFrame` | None
            GeoDataFrame with county level setback regulations merged
            with county geometries, use for intersecting with setback
            features.
        Multiplier : float | None
            Generic setbacks multiplier
        """
        if regulations_fpath:
            if multiplier:
                msg = ('A regulation .csv file was also provided and '
                       'will be used to determine setback multipliers!')
                logger.warning(msg)
                warn(msg)

            multiplier = None
            regulations = self._parse_regulations(regulations_fpath)
            logger.debug('Computing setbacks using regulations provided in: {}'
                         .format(regulations_fpath))
        elif multiplier:
            regulations = None
            logger.debug('Computing setbacks using generic plant height '
                         'multiplier of {}'.format(multiplier))
        else:
            msg = ('Computing setbacks requires either a regulations '
                   '.csv file or a generic multiplier!')
            logger.error(msg)
            raise RuntimeError(msg)

        return regulations, multiplier

    def _parse_regulations(self, regulations_fpath):
        """Parse regulations file.

        Parameters
        ----------
        regulations_fpath : str
            Path to regulations .csv file.

        Returns
        -------
        regulations: `geopandas.GeoDataFrame`
            GeoDataFrame with county level setback regulations merged
            with county geometries, use for intersecting with setback
            features.
        """
        try:
            regulations = parse_table(regulations_fpath)
            regulations = self._parse_county_regulations(regulations)
            log_mem(logger)

            out_path = regulations_fpath.split('.')[0] + '.gpkg'
            logger.debug('Saving regulations with county geometries as: '
                         '{}'.format(out_path))
            regulations.to_file(out_path, driver='GPKG')
        except ValueError:
            regulations = gpd.read_file(regulations_fpath)

        fips_check = regulations['geometry'].isnull()
        if fips_check.any():
            msg = ('The following county FIPS were requested in the '
                   'regulations but were not available in the '
                   'Exclusions "cnty_fips" layer:\n{}'
                   .format(regulations.loc[fips_check, 'FIPS']))
            logger.error(msg)
            raise RuntimeError(msg)

        return regulations.to_crs(crs=self.crs)

    def _parse_county_regulations(self, regulations):
        """Parse the county regulations.

        Parse regulations, combine with county geometries from
        exclusions .h5 file. The county geometries are intersected with
        features to compute county specific setbacks.

        Parameters
        ----------
        regulations : pandas.DataFrame
            Regulations table

        Returns
        -------
        regulations: `geopandas.GeoDataFrame`
            GeoDataFrame with county level setback regulations merged
            with county geometries, use for intersecting with setback
            features.
        """
        if 'FIPS' not in regulations:
            msg = ('Regulations does not have county FIPS! Please add a '
                   '"FIPS" columns with the unique county FIPS values.')
            logger.error(msg)
            raise RuntimeError(msg)

        if 'geometry' not in regulations:
            regulations['geometry'] = None

        regulations = regulations.set_index('FIPS')

        logger.info('Merging county geometries w/ local regulations')
        with ExclusionLayers(self._excl_fpath) as exc:
            fips = exc['cnty_fips']
            profile = exc.get_layer_profile('cnty_fips')

        s = features.shapes(
            fips.astype(np.int32),
            transform=profile['transform']
        )
        for p, v in s:
            v = int(v)
            if v in regulations.index:
                regulations.at[v, 'geometry'] = shape(p)

        regulations = gpd.GeoDataFrame(
            regulations, crs=self.crs, geometry='geometry'
        )

        return regulations.reset_index()

    @property
    def base_setback_dist(self):
        """The base setback distance, in meters.

        Returns
        -------
        float
        """
        return self._base_setback_dist

    @property
    def generic_setback(self):
        """Default setback of plant height * multiplier.

        This value is used for global setbacks.

        Returns
        -------
        float
        """
        if self.multiplier is None:
            setback = None
        else:
            setback = self.base_setback_dist * self.multiplier

        return setback

    @property
    def multiplier(self):
        """Generic setback multiplier.

        Returns
        -------
        int | float
        """
        return self._multi

    @property
    def arr_shape(self):
        """Rasterize array shape.

        Returns
        -------
        tuple
        """
        return self._shape

    @property
    def profile(self):
        """Geotiff profile.

        Returns
        -------
        dict
        """
        return self._profile

    @property
    def crs(self):
        """Coordinate reference system.

        Returns
        -------
        str
        """
        return self.profile['crs']

    @property
    def regulations(self):
        """Regulations table.

        Returns
        -------
        geopandas.GeoDataFrame | None
        """
        return self._regulations

    def _parse_features(self, features_fpath):
        """Abstract method to parse features.

        Parameters
        ----------
        features_fpath : str
            Path to file containing features to setback from.

        Returns
        -------
        `geopandas.GeoDataFrame`
            Geometries of features to setback from in exclusion
            coordinate system.
        """
        return gpd.read_file(features_fpath).to_crs(crs=self.crs)

    # pylint: disable=unused-argument
    def _check_regulations(self, features_fpath):
        """Reduce regulations to state corresponding to features_fpath.

        Parameters
        ----------
        features_fpath : str
            Path to shape file with features to compute setbacks from.

        Returns
        -------
        regulations : geopandas.GeoDataFrame | None
            Regulations table.
        """
        logger.debug('Computing setbacks for regulations in {} counties'
                     .format(len(self.regulations)))

        return self.regulations

    @abstractmethod
    def get_regulation_setback(self, county_regulations):
        """Compute the setback distance for the county.

        Compute the setback distance (in meters) from the
        county regulations or the plant height.

        Parameters
        ----------
        county_regulations : pandas.Series
            Pandas Series with regulations for a single county
            or feature type. At a minimum, this Series must
            contain the following columns: `Value Type`, which
            specifies wether the value is a multiplier or static height,
            `Value`, which specifies the numeric value of the setback or
            multiplier. Valid options for the `Value Type` are:
                - "Max-tip Height Multiplier"
                - "Rotor-Diameter Multiplier"
                - "Hub-height Multiplier"
                - "Structure Height multiplier"
                - "Meters"

        Returns
        -------
        setback : float | None
            Setback distance in meters, or `None` if the setback
            `Value Type` was not recognized.
        """

    @staticmethod
    def _compute_local_setbacks(features, cnty, setback):
        """Compute local features setbacks.

        This method will compute the setbacks using a county-specific
        regulations file that specifies either a static setback or a
        multiplier value that will be used along with plant height
        specifications to compute the setback.

        Parameters
        ----------
        features : geopandas.GeoDataFrame
            Features to setback from.
        cnty : geopandas.GeoDataFrame
            Regulations for a single county.
        setback : int
            Setback distance in meters.

        Returns
        -------
        setbacks : list
            List of setback geometries.
        """
        logger.debug('- Computing setbacks for county FIPS {}'
                     .format(cnty.iloc[0]['FIPS']))
        log_mem(logger)
        mask = features.centroid.within(cnty['geometry'].values[0])
        tmp = features.loc[mask]
        tmp.loc[:, 'geometry'] = tmp.buffer(setback)

        setbacks = [(geom, 1) for geom in tmp['geometry']]

        return setbacks

    def _no_exclusions_array(self):
        """Get an array of the correct shape reprenting no exclusions.

        The array contains all zeros, and a new one is created
        for every function call.

        Returns
        -------
        np.array
            Array of zeros representing no exclusions.
        """
        return np.zeros(self.arr_shape[1:], dtype='uint8')

    def _rasterize_setbacks(self, shapes):
        """Convert setbacks geometries into exclusions array.

        Parameters
        ----------
        shapes : list, optional
            List of (geometry, 1) pairs to rasterize. Each geometry is a
            feature buffered by the desired setback distance in meters.
            If `None` or empty list, returns array of zeros.

        Returns
        -------
        arr : ndarray
            Rasterized array of setbacks.
        """
        logger.debug('Generating setbacks exclusion array of shape {}'
                     .format(self.arr_shape))
        log_mem(logger)
        arr = self._no_exclusions_array()
        if shapes:
            features.rasterize(shapes=shapes,
                               out=arr,
                               out_shape=self.arr_shape[1:],
                               fill=0,
                               transform=self.profile['transform'])

        return arr

    def _write_setbacks(self, geotiff, setbacks, replace=False):
        """
        Write setbacks to geotiff, replace if requested

        Parameters
        ----------
        geotiff : str
            Path to geotiff file to save setbacks too
        setbacks : ndarray
            Rasterized array of setbacks
        replace : bool, optional
            Flag to replace local layer data with arr if layer already
            exists in the exclusion .h5 file. By default `False`.
        """
        if os.path.exists(geotiff):
            if not replace:
                msg = ('{} already exists. To replace it set "replace=True"'
                       .format(geotiff))
                logger.error(msg)
                raise IOError(msg)
            else:
                msg = ('{} already exists and will be replaced!'
                       .format(geotiff))
                logger.warning(msg)
                warn(msg)

        ExclusionsConverter._write_geotiff(geotiff, self.profile, setbacks)

    def compute_local_setbacks(self, features_fpath, max_workers=None):
        """Compute local setbacks for all counties either.

        Parameters
        ----------
        features_fpath : str
            Path to shape file with features to compute setbacks from
        max_workers : int, optional
            Number of workers to use for setback computation, if 1 run
            in serial, if > 1 run in parallel with that many workers,
            if `None` run in parallel on all available cores.
            By default `None`.

        Returns
        -------
        setbacks : ndarray
            Raster array of setbacks.
        """
        regulations = self._check_regulations(features_fpath)
        if regulations.empty:
            return self._no_exclusions_array()

        setbacks = []
        setback_features = self._parse_features(features_fpath)
        if max_workers is None:
            max_workers = os.cpu_count()

        log_mem(logger)
        if max_workers > 1:
            logger.info('Computing local setbacks in parallel using {} '
                        'workers'.format(max_workers))
            loggers = [__name__, 'reVX']
            with SpawnProcessPool(max_workers=max_workers,
                                  loggers=loggers) as exe:
                futures = []
                for i in range(len(regulations)):
                    cnty = regulations.iloc[[i]].copy()
                    setback = self.get_regulation_setback(cnty.iloc[0])
                    if setback is not None:
                        idx = setback_features.sindex.intersection(
                            cnty.total_bounds
                        )
                        cnty_feats = setback_features.iloc[list(idx)].copy()
                        future = exe.submit(self._compute_local_setbacks,
                                            cnty_feats, cnty, setback)
                        futures.append(future)

                for i, future in enumerate(as_completed(futures)):
                    setbacks.extend(future.result())
                    logger.debug('Computed setbacks for {} of {} counties'
                                 .format((i + 1), len(regulations)))
        else:
            logger.info('Computing local setbacks in serial')
            for i in range(len(regulations)):
                cnty = regulations.iloc[[i]]
                setback = self.get_regulation_setback(cnty.iloc[0])
                if setback is not None:
                    idx = setback_features.sindex.intersection(
                        cnty.total_bounds
                    )
                    cnty_feats = setback_features.iloc[list(idx)]
                    setbacks.extend(self._compute_local_setbacks(cnty_feats,
                                                                 cnty,
                                                                 setback))
                    logger.debug('Computed setbacks for {} of {} counties'
                                 .format((i + 1), len(regulations)))

        return self._rasterize_setbacks(setbacks)

    def compute_generic_setbacks(self, features_fpath):
        """Compute generic setbacks.

        This method will compute the setbacks using a generic setback
        of `base_setback_dist * multiplier`.

        Parameters
        ----------
        features_fpath : str
            Path to shape file with features to compute setbacks from.

        Returns
        -------
        setbacks : ndarray
            Raster array of setbacks
        """
        logger.info('Computing generic setbacks')
        setback_features = self._parse_features(features_fpath)
        setback_features.loc[:, 'geometry'] = setback_features.buffer(
            self.generic_setback
        )
        setbacks = [(geom, 1) for geom in setback_features['geometry']]

        return self._rasterize_setbacks(setbacks)

    def compute_setbacks(self, features_fpath, max_workers=None,
                         geotiff=None, replace=False):
        """
        Compute setbacks for all states either in serial or parallel.
        Existing setbacks are computed if a regulations file was
        supplied during class initialization, otherwise generic setbacks
        are computed.

        Parameters
        ----------
        features_fpath : str
            Path to shape file with features to compute setbacks from
        max_workers : int, optional
            Number of workers to use for setback computation, if 1 run
            in serial, if > 1 run in parallel with that many workers,
            if `None`, run in parallel on all available cores.
            By default `None`.
        geotiff : str, optional
            Path to save geotiff containing rasterized setbacks.
            By default `None`.
        replace : bool, optional
            Flag to replace geotiff if it already exists.
            By default `False`.

        Returns
        -------
        setbacks : ndarray
            Raster array of setbacks
        """
        if self._regulations is not None:
            setbacks = self.compute_local_setbacks(features_fpath,
                                                   max_workers=max_workers)
        else:
            setbacks = self.compute_generic_setbacks(features_fpath)

        if geotiff is not None:
            logger.debug('Writing setbacks to {}'.format(geotiff))
            self._write_setbacks(geotiff, setbacks, replace=replace)

        return setbacks

    @staticmethod
    def _get_feature_paths(features_fpath):
        """Ensure features path exists and return as list.

        Parameters
        ----------
        features_fpath : str
            Path to features file. This path can contain
            any pattern that can be used in the glob function.
            For example, `/path/to/features/[A]*` would match
            with all the features in the direcotry
            `/path/to/features/` that start with "A". This input
            can also be a directory, but that directory must ONLY
            contain feature files. If your feature files are mixed
            with other files or directories, use something like
            `/path/to/features/*.geojson`.

        Returns
        -------
        features_fpath : list
            Features path as a list of strings.

        Notes
        -----
        This method is required for `run` classmethods for
        feature setbacks that are spread out over multiple
        files.
        """
        glob_path = pathlib.Path(features_fpath)
        if glob_path.is_dir():
            glob_path = glob_path / '*'

        paths = [str(f) for f in glob_path.parent.glob(glob_path.name)]
        if not paths:
            msg = 'No files found matching the input {!r}!'
            msg = msg.format(features_fpath)
            logger.error(msg)
            raise FileNotFoundError(msg)

        return paths



# -*- coding: utf-8 -*-
"""
reVX Offshore Sub-package Configurations
"""

from reV.config.base_analysis_config import AnalysisConfig


class AssemblyAreasConfig(AnalysisConfig):
    """Config framework for assembly area distance calculation"""

    NAME = 'AssemblyAreas'
    REQUIREMENTS = ('assembly_areas', 'excl_fpath')

    def __init__(self, config):
        """
        Parameters
        ----------
        config : str | dict
            Path to config .json or pre-extracted config input dictionary.
        """
        super().__init__(config)
        self._default_ports_dset = 'ports_construction_nolimits'

    @property
    def assembly_areas(self):
        """
        Path to csv or json file containing assembly area locations. If
        provided compute distance from ports to assembly areas and save as
        a table to excl_fpath (requried)
        """
        return self['assembly_areas']

    @property
    def excl_fpath(self):
        """Get the exclusions .h5 file path (required)."""
        return self['excl_fpath']

    @property
    def ports_dset(self):
        """
        Distance to ports layer/dataset name in excl_fpath, by default
        'ports_construction_nolimits'
        """
        dset = self.get('ports_dsets',
                        self._default_ports_dset)

        return dset


class DistToPortsConfig(AnalysisConfig):
    """Config framework for distance to port calculation"""

    NAME = 'DistToPorts'
    REQUIREMENTS = ('ports_fpath', 'excl_fpath')

    def __init__(self, config):
        """
        Parameters
        ----------
        config : str | dict
            Path to config .json or pre-extracted config input dictionary.
        """
        super().__init__(config)
        self._default_dist_layer = 'dist_to_coast'
        self._default_replace = False

    @property
    def ports_fpath(self):
        """
        Path to shape file containing ports to compute least cost distance
        to (required).
        """
        return self['ports_fpath']

    @property
    def excl_fpath(self):
        """Get the exclusions .h5 file path (required)."""
        return self['excl_fpath']

    @property
    def input_dist_layer(self):
        """
        Exclusions layer with distance to shore. Only used if
        'dist_to_coast' is a .h5 exclusions file path.
        """
        return self.get('input_dist_layer', self._default_dist_layer)

    @property
    def replace(self):
        """
        Flag to replace existing ports geotiffs, by default False
        """
        return self.get('replace', self._default_replace)


class OffshoreInputsConfig(AnalysisConfig):
    """Config framework for extraction offshore inputs"""

    NAME = 'OffshoreInputs'
    REQUIREMENTS = ('inputs_fpath', 'offshore_sites', 'input_layers')

    def __init__(self, config):
        """
        Parameters
        ----------
        config : str | dict
            Path to config .json or pre-extracted config input dictionary.
        """
        super().__init__(config)
        self._default_tm_dset = 'techmap_wtk'

    @property
    def inputs_fpath(self):
        """
        Path to offshore inputs .h5 file (required).
        """
        return self['inputs_fpath']

    @property
    def offshore_sites(self):
        """
        - Path to .csv|.json file with offshore sites meta data
        - Path to a WIND Toolkit .h5 file to extact site meta from
        - List, tuple, or vector of offshore gids
        - Pre-extracted site meta DataFrame
        (required)
        """
        return self['offshore_sites']

    @property
    def input_layers(self):
        """
        Path to json file containing 'input_layers', list of input layers, to
        extract, or dictionary mapping the input layers to extract (keys) to
        the column names to save them under (values) (required)
        """
        return self['input_layers']

    @property
    def tm_dset(self):
        """
        Dataset / layer name for wind toolkit techmap,
        by default 'techmap_wtk'
        """
        return self.get('tm_dset', self._default_tm_dset)



# -*- coding: utf-8 -*-
"""
PLEXOS Plants command line interface (cli).
"""
import click
import logging
import os

from rex.utilities.cli_dtypes import STR, INT
from rex.utilities.hpc import SLURM
from rex.utilities.loggers import init_mult

from reVX.plexos.plexos_plants import PlantProfileAggregation
from reVX import __version__

logger = logging.getLogger(__name__)


@click.group(invoke_without_command=True)
@click.version_option(version=__version__)
@click.option('--name', '-n', default='plexos-plants', type=STR,
              help='Job name. Default is "plexos-plants".')
@click.option('--plexos_table', '-pt', required=True,
              type=click.Path(exists=True),
              help=('Path to PLEXOS table of bus locations and capacity .csv'))
@click.option('--sc_table', '-st', required=True,
              type=click.Path(exists=True),
              help=('Path to Supply Curve table .csv'))
@click.option('--cf_fpath', '-cf', required=True,
              type=click.Path(exists=True),
              help=('Path to reV Generation output .h5 file'))
@click.option('--out_dir', '-out', required=True, type=click.Path(),
              help='Directory to dump output files')
@click.option('--dist_percentile', '-dp', type=int, default=90,
              show_default=True,
              help=('Percentile to use to compute distance threshold using '
                    'sc_gid to SubStation distance, by default 90'))
@click.option('--lcoe_col', '-lc', type=str, default='total_lcoe',
              show_default=True,
              help="LCOE column to sort by, by default 'total_lcoe'")
@click.option('--lcoe_thresh', '-lt', type=float, default=1.3,
              show_default=True,
              help=('LCOE threshold multiplier, exclude sc_gids above '
                    'threshold, by default 1.3'))
@click.option('--max_workers', '-mw', type=INT, default=None,
              show_default=True,
              help=('Number of workers to use for point and plant creation, '
                    '1 == serial, > 1 == parallel, None == parallel using all '
                    'available cpus, by default None'))
@click.option('--points_per_worker', type=int, default=400,
              show_default=True,
              help='Number of points to create on each worker, by default 400')
@click.option('--plants_per_worker', type=int, default=40,
              show_default=True,
              help=('Number of plants to identify on each worker, by default '
                    '40'))
@click.option('-o', '--offshore', is_flag=True,
              help='Include offshore points, by default False')
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def main(ctx, name, plexos_table, sc_table, cf_fpath, out_dir, dist_percentile,
         lcoe_col, lcoe_thresh, max_workers, points_per_worker,
         plants_per_worker, offshore, verbose):
    """PLEXOS plant Command Line Interface"""

    ctx.ensure_object(dict)
    ctx.obj['NAME'] = name
    ctx.obj['PLEXOS_TABLE'] = plexos_table
    ctx.obj['SC_TABLE'] = sc_table
    ctx.obj['CF_FPATH'] = cf_fpath
    ctx.obj['OUT_DIR'] = out_dir
    ctx.obj['DIST_PERCENTILE'] = dist_percentile
    ctx.obj['LCOE_COL'] = lcoe_col
    ctx.obj['LCOE_THRESH'] = lcoe_thresh
    ctx.obj['MAX_WORKERS'] = max_workers
    ctx.obj['POINTS_PER_WORKER'] = points_per_worker
    ctx.obj['PLANTS_PER_WORKER'] = plants_per_worker
    ctx.obj['OFFSHORE'] = offshore
    ctx.obj['VERBOSE'] = verbose

    if ctx.invoked_subcommand is None:
        log_modules = [__name__, 'reVX', 'reV', 'rex']
        init_mult(name, out_dir, modules=log_modules,
                  verbose=verbose)
        logger.info('Aggregating Plant for buses in PLEXOS table: {}'
                    .format(plexos_table))
        out_fpath = '-' + os.path.basename(cf_fpath).split('_')[-1]
        out_fpath = os.path.basename(plexos_table).replace('.csv', out_fpath)
        out_fpath = os.path.join(out_dir, out_fpath)
        logger.info('Saving Aggregated Plant Profiles to {}'
                    .format(out_fpath))
        PlantProfileAggregation.run(plexos_table, sc_table, cf_fpath,
                                    out_fpath, dist_percentile=dist_percentile,
                                    lcoe_col=lcoe_col, lcoe_thresh=lcoe_thresh,
                                    max_workers=max_workers,
                                    points_per_worker=points_per_worker,
                                    plants_per_worker=plants_per_worker,
                                    offshore=offshore)


def get_node_cmd(name, plexos_table, sc_table, cf_fpath, out_dir,
                 dist_percentile, lcoe_col, lcoe_thresh, max_workers,
                 points_per_worker, plants_per_worker, offshore, verbose):
    """Build PLEXOS Plant CLI command."""

    args = ['-n {}'.format(SLURM.s(name)),
            '-pt {}'.format(SLURM.s(plexos_table)),
            '-sc {}'.format(SLURM.s(sc_table)),
            '-cf {}'.format(SLURM.s(cf_fpath)),
            '-out {}'.format(SLURM.s(out_dir)),
            '-dp {}'.format(SLURM.s(dist_percentile)),
            '-lc {}'.format(SLURM.s(lcoe_col)),
            '-lt {}'.format(SLURM.s(lcoe_thresh)),
            '-mw {}'.format(SLURM.s(max_workers)),
            '--points_per_worker={}'.format(SLURM.s(points_per_worker)),
            '--plants_per_worker={}'.format(SLURM.s(plants_per_worker)),
            ]

    if offshore:
        args.append('-o')

    if verbose:
        args.append('-v')

    cmd = 'python -m reVX.plexos.plexos_plants_cli {}'.format(' '.join(args))
    logger.debug('Creating the following command line call:\n\t{}'.format(cmd))

    return cmd


@main.command()
@click.option('--alloc', '-a', required=True, type=STR,
              help='Eagle allocation account name.')
@click.option('--walltime', '-wt', default=1.0, type=float,
              show_default=True,
              help='Eagle walltime request in hours. Default is 1.0')
@click.option('--feature', '-l', default=None, type=STR,
              show_default=True,
              help=('Additional flags for SLURM job. Format is "--qos=high" '
                    'or "--depend=[state:job_id]". Default is None.'))
@click.option('--stdout_path', '-sout', default=None, type=STR,
              show_default=True,
              help='Subprocess standard output path. Default is in out_dir.')
@click.pass_context
def eagle(ctx, alloc, walltime, feature, stdout_path):
    """Eagle submission tool for PLEXOS aggregation."""

    name = ctx.obj['NAME']
    plexos_table = ctx.obj['PLEXOS_TABLE']
    sc_table = ctx.obj['SC_TABLE']
    cf_fpath = ctx.obj['CF_FPATH']
    out_dir = ctx.obj['OUT_DIR']
    dist_percentile = ctx.obj['DIST_PERCENTILE']
    lcoe_col = ctx.obj['LCOE_COL']
    lcoe_thresh = ctx.obj['LCOE_THRESH']
    max_workers = ctx.obj['MAX_WORKERS']
    points_per_worker = ctx.obj['POINTS_PER_WORKER']
    plants_per_worker = ctx.obj['PLANTS_PER_WORKER']
    offshore = ctx.obj['OFFSHORE']
    verbose = ctx.obj['VERBOSE']

    if stdout_path is None:
        stdout_path = os.path.join(out_dir, 'stdout/')

    slurm_manager = ctx.obj.get('SLURM_MANAGER', None)
    if slurm_manager is None:
        slurm_manager = SLURM()
        ctx.obj['SLURM_MANAGER'] = slurm_manager

    cmd = get_node_cmd(name, plexos_table, sc_table, cf_fpath, out_dir,
                       dist_percentile, lcoe_col, lcoe_thresh, max_workers,
                       points_per_worker, plants_per_worker, offshore, verbose)

    logger.info('Running reVX plexos plant aggregation on Eagle with '
                'node name "{}"'.format(name))

    out = slurm_manager.sbatch(cmd, alloc=alloc, walltime=walltime,
                               feature=feature, name=name,
                               stdout_path=stdout_path)[0]
    if out:
        msg = ('Kicked off reVX plexos aggregation job "{}" '
               '(SLURM jobid #{}) on Eagle.'
               .format(name, out))
    else:
        msg = ('Was unable to kick off reVX plexos aggregation job "{}". '
               'Please see the stdout error messages'
               .format(name))

    click.echo(msg)
    logger.info(msg)


if __name__ == '__main__':
    try:
        main(obj={})
    except Exception:
        logger.exception('Error running PLEXOS-Plant CLI')
        raise



# -*- coding: utf-8 -*-
"""
reVX Turbine Flicker Configuration
"""

from reV.config.base_analysis_config import AnalysisConfig


class TurbineFlickerConfig(AnalysisConfig):
    """Config framework for turbine flicker calculation"""

    NAME = 'TurbineFlicker'
    REQUIREMENTS = ('excl_fpath', 'res_fpath', 'building_layer', 'hub_height')

    def __init__(self, config):
        """
        Parameters
        ----------
        config : str | dict
            Path to config .json or pre-extracted config input dictionary.
        """
        super().__init__(config)
        self._default_tm_dset = 'techmap_wtk'
        self._default_resolution = 128
        self._default_building_threshold = 0
        self._default_flicker_threshold = 30

    @property
    def excl_fpath(self):
        """Get the exclusions .h5 file path (required)."""
        return self['excl_fpath']

    @property
    def res_fpath(self):
        """Get the resource .h5 file path (required)."""
        return self['res_fpath']

    @property
    def building_layer(self):
        """Get the building layer name."""
        return self['building_layer']

    @property
    def hub_height(self):
        """
        Get the turbine hub-height for which shadow flicker will be computed.
        """
        return self['hub_height']

    @property
    def tm_dset(self):
        """Get the techmap dataset name."""
        return self.get('tm_dset', self._default_tm_dset)

    @property
    def resolution(self):
        """Get the supply curve resolution."""
        return self.get('resolution', self._default_resolution)

    @property
    def building_threshold(self):
        """
        Get the threshold for which buildings are identified in the
        building_layer.
        """
        return self.get('building_threshold', self._default_building_threshold)

    @property
    def flicker_threshold(self):
        """
        Get the threshold at which shadow flicker will lead to an exclusions,
        values are in hours
        """
        return self.get('flicker_threshold', self._default_flicker_threshold)

    @property
    def out_layer(self):
        """
        Get the output layer name under which turbine flicker exclusions will
        be saved
        """
        return self.get('out_layer', None)



# -*- coding: utf-8 -*-
"""
Module to compute least cost xmission paths, distances, and costs one or
more SC points
"""
from concurrent.futures import as_completed
import geopandas as gpd
import json
import logging
import numpy as np
import os
import pandas as pd
from pyproj.crs import CRS
import rasterio
from scipy.spatial import cKDTree
from shapely.geometry import Point
import time

from reV.handlers.exclusions import ExclusionLayers
from reV.supply_curve.extent import SupplyCurveExtent
from rex.utilities.execution import SpawnProcessPool
from rex.utilities.loggers import log_mem

from reVX.least_cost_xmission.config import (TRANS_LINE_CAT, LOAD_CENTER_CAT,
                                             SINK_CAT, SUBSTATION_CAT)
from reVX.least_cost_xmission.least_cost_paths import LeastCostPaths
from reVX.least_cost_xmission.trans_cap_costs import TransCapCosts

logger = logging.getLogger(__name__)


class LeastCostXmission(LeastCostPaths):
    """
    Compute Least Cost tie-line paths and full transmission cap cost
    for all possible connections to all supply curve points
    -
    """
    REQUIRED_LAYRES = ['transmission_barrier', 'ISO_regions']

    def __init__(self, cost_fpath, features_fpath, resolution=128,
                 xmission_config=None):
        """
        Parameters
        ----------
        cost_fpath : str
            Path to h5 file with cost rasters and other required layers
        features_fpath : str
            Path to geopackage with transmission features
        resolution : int, optional
            SC point resolution, by default 128
        xmission_config : str | dict | XmissionConfig, optional
            Path to Xmission config .json, dictionary of Xmission config
            .jsons, or preloaded XmissionConfig objects, by default None
        """
        self._check_layers(cost_fpath)
        self._config = TransCapCosts._parse_config(
            xmission_config=xmission_config)

        (self._sc_points, self._features,
         self._sub_lines_mapping, self._shape) =\
            self._map_to_costs(cost_fpath, features_fpath,
                               resolution=resolution)
        self._cost_fpath = cost_fpath
        self._tree = None
        self._sink_coords = None
        self._min_line_len = (resolution * 0.09) / 2

        logger.debug('{} initialized'.format(self))

    def __repr__(self):
        msg = ("{} to be computed for {} sc_points and {} features"
               .format(self.__class__.__name__,
                       len(self.sc_points),
                       len(self.features)))

        return msg

    @property
    def sc_points(self):
        """
        Table of supply curve points

        Returns
        -------
        gpd.GeoDataFrame
        """
        return self._sc_points

    @property
    def features(self):
        """
        Table of features to compute paths for

        Returns
        -------
        pandas.DataFrame
        """
        return self._features

    @property
    def sub_lines_mapping(self):
        """
        Series mapping substations  to the transmission lines connected
        to each substation

        Returns
        -------
        pandas.Series
        """
        return self._sub_lines_mapping

    @property
    def sink_coords(self):
        """
        Inf sink coordinates (row, col)

        Returns
        -------
        ndarray
        """
        if self._sink_coords is None:
            mask = self.features['category'] == SINK_CAT
            self._sink_coords = self.features.loc[mask, ['row', 'col']].values

        return self._sink_coords

    @property
    def sink_tree(self):
        """
        cKDTree for infinite sinks

        Returns
        -------
        cKDTree
        """
        if self._tree is None:
            self._tree = cKDTree(self.sink_coords)

        return self._tree

    @staticmethod
    def _load_trans_feats(features_fpath):
        """
        Load existing transmission features from disk. Substations will be
        loaded from cache file if it exists

        Parameters
        ----------
        features_fpath : str
            Path to geopackage with trans features

        Returns
        -------
        features : gpd.GeoDataFrame
            DataFrame of transmission features
        sub_line_map : pandas.Series
            Mapping of sub-station trans_gid to connected tranmission line
            trans_gids
        """
        logger.debug('Loading transmission features')
        features = gpd.read_file(features_fpath)
        features = features.drop(columns=['bgid', 'egid', 'cap_left'],
                                 errors='ignore')
        mapping = {'gid': 'trans_gid', 'trans_gids': 'trans_line_gids'}
        features = features.rename(columns=mapping)

        features['min_volts'] = 0
        features['max_volts'] = 0

        # Transmission lines
        mask = features['category'] == TRANS_LINE_CAT
        voltage = features.loc[mask, 'voltage'].values
        features.loc[mask, 'min_volts'] = voltage
        features.loc[mask, 'max_volts'] = voltage

        # Load Center and Sinks
        mask = features['category'].isin([LOAD_CENTER_CAT, SINK_CAT])
        features.loc[mask, 'min_volts'] = 1
        features.loc[mask, 'max_volts'] = 9999

        sub_lines_map = {}
        mask = features['category'] == SUBSTATION_CAT
        bad_subs = np.zeros(len(features), dtype=bool)
        for idx, row in features.loc[mask].iterrows():
            gid = row['trans_gid']
            lines = row['trans_line_gids']
            if isinstance(lines, str):
                lines = json.loads(lines)

            sub_lines_map[gid] = lines
            lines_mask = features['trans_gid'].isin(lines)
            voltage = features.loc[lines_mask, 'voltage'].values

            if np.max(voltage) >= 69:
                features.loc[idx, 'min_volts'] = np.min(voltage)
                features.loc[idx, 'max_volts'] = np.max(voltage)
            else:
                bad_subs[idx] = True

        if any(bad_subs):
            msg = ("The following sub-stations do not have the minimum "
                   "required voltage of 69 kV and will be dropped:\n{}"
                   .format(features.loc[bad_subs, 'trans_gid']))
            logger.warning(msg)
            features = features.loc[~bad_subs].reset_index(drop=True)

        return features, pd.Series(sub_lines_map)

    @staticmethod
    def _create_sc_points(cost_fpath, resolution=128):
        """
        Load SC points, covert row/col to array wide, and determine x/y for
        reV projection

        Parameters
        ----------
        cost_fpath : str
            Path to h5 file with cost rasters and other required layers
        resolution : int, optional
            SC point resolution, by default 128

        Returns
        sc_points : gpd.GeoDataFrame
            SC points
        """
        logger.debug('Loading Supply Curve Points')
        sce = SupplyCurveExtent(cost_fpath, resolution=resolution)
        sc_points = sce.points.rename(columns={'row_ind': 'sc_row_ind',
                                               'col_ind': 'sc_col_ind'})
        shape = sce.excl_shape
        sc_points['sc_point_gid'] = sc_points.index.values

        row = np.round(sc_points['sc_row_ind'] * resolution + resolution / 2)
        row = np.where(row >= shape[0], shape[0] - 1, row)
        sc_points['row'] = row.astype(int)

        col = np.round(sc_points['sc_col_ind'] * resolution + resolution / 2)
        col = np.where(col >= shape[1], shape[1] - 1, col)
        sc_points['col'] = col.astype(int)

        return sc_points

    @staticmethod
    def _get_feature_cost_indices(features, crs, transform, shape):
        """
        Map features to cost row, col indicies using rasterio transform

        Parameters
        ----------
        features : gpd.GeoDataFrame
            GeoDataFrame of features to map to cost raster
        crs : pyproj.crs.CRS
            CRS of cost raster
        transform : raster.Affine
            Transform of cost raster
        shape : tuple
            Cost raster shape

        Returns
        -------
        row : ndarray
            Vector of row indicies for each feature
        col : ndarray
            Vector of col indicies for each features
        mask : ndarray
            Boolean mask of features with indicies outside of cost raster
        """
        row, col, mask = super(LeastCostXmission,
                               LeastCostXmission)._get_feature_cost_indices(
            features, crs, transform, shape)

        t_lines = features['category'] == TRANS_LINE_CAT
        mask |= t_lines

        row[t_lines] = np.where(row[t_lines] >= 0, row[t_lines], 0)
        row[t_lines] = np.where(row[t_lines] < shape[0], row[t_lines],
                                shape[0] - 1)
        col[t_lines] = np.where(col[t_lines] >= 0, col[t_lines], 0)
        col[t_lines] = np.where(col[t_lines] < shape[1], col[t_lines],
                                shape[1] - 1)

        return row, col, mask

    @classmethod
    def _map_to_costs(cls, cost_fpath, features_fpath, resolution=128):
        """
        Map supply curve points and transmission features to cost array pixel
        indices

        Parameters
        ----------
        cost_fpath : str
            Path to h5 file with cost rasters and other required layers
        features_fpath : str
            Path to geopackage with transmission features
        resolution : int, optional
            SC point resolution, by default 128

        Returns
        -------
        sc_point : gpd.GeoDataFrame
            Table of supply curve points to connect to tranmission
        features : gpd.GeoDataFrame
            Table of transmission features
        sub_lines_map : pandas.Series
            Series mapping substations  to the transmission lines connected
            to each substation
        """
        with ExclusionLayers(cost_fpath) as f:
            crs = CRS.from_string(f.crs)
            transform = rasterio.Affine(*f.profile['transform'])
            shape = f.shape
            regions = f['ISO_regions']

        features, sub_lines_map = cls._load_trans_feats(features_fpath)
        row, col, mask = cls._get_feature_cost_indices(features, crs,
                                                       transform, shape)
        if any(~mask):
            msg = ("The following features are outside of the cost exclusion "
                   "domain and will be dropped:\n{}"
                   .format(features.loc[~mask, 'trans_gid']))
            logger.warning(msg)
            row = row[mask]
            col = col[mask]
            features = features.loc[mask].reset_index(drop=True)

        features['row'] = row
        features['col'] = col
        features['region'] = regions[row, col]

        logger.debug('Converting SC points to GeoDataFrame')
        sc_points = cls._create_sc_points(cost_fpath, resolution=resolution)
        x, y = rasterio.transform.xy(transform, sc_points['row'].values,
                                     sc_points['col'].values)
        geo = [Point(xy) for xy in zip(x, y)]
        sc_points = gpd.GeoDataFrame(sc_points, crs=features.crs,
                                     geometry=geo)

        return sc_points, features, sub_lines_map, shape

    def _clip_to_sc_point(self, sc_point, tie_line_voltage, nn_sinks=2,
                          clipping_buffer=1.05):
        """
        Clip costs raster to AOI around SC point, and get substations,
        load centers, and sinks within the clipped region.

        Parameters
        ----------
        sc_point : gpd.GeoSeries
            SC point to clip raster around
        nn_sinks : int, optional
            Number of nearest neighbor sinks to clip to
        clipping_buffer : float, optional
            Buffer to increase clipping radius by, by default 1.05

        Returns
        -------
        radius : int
            Clipping radius in cost raster pixels
        x_feats : pd.DataFrame
            Substatations, load centers, sinks, and nearest points on t-lines
            to SC point
        """
        logger.debug('Clipping features to sc_point {}'.format(sc_point.name))
        if len(self.sink_coords) > 2:
            row, col = sc_point[['row', 'col']].values
            _, pos = self.sink_tree.query([row, col], k=nn_sinks)
            radius = np.abs(self.sink_coords[pos] - np.array([row, col])).max()
            radius = int(np.ceil(radius * clipping_buffer))

            logger.debug('Radius to {} nearest sink is: {}'
                         .format(nn_sinks, radius))
            row_min = max(row - radius, 0)
            row_max = min(row + radius, self._shape[0])
            col_min = max(col - radius, 0)
            col_max = min(col + radius, self._shape[1])
            logger.debug('Extracting all transmission features in the row '
                         'slice {}:{} and column slice {}:{}'
                         .format(row_min, row_max, col_min, col_max))

            # Clip transmission features
            mask = self.features['row'] >= row_min
            mask &= self.features['row'] < row_max
            mask &= self.features['col'] >= col_min
            mask &= self.features['col'] < col_max
            sc_features = self.features.loc[mask].copy(deep=True)
            logger.debug('{} transmission features found in clipped area with '
                         'radius {}'
                         .format(len(sc_features), radius))
        else:
            radius = None
            sc_features = self.features.copy(deep=True)

        mask = self.features['max_volts'] >= tie_line_voltage
        sc_features = sc_features.loc[mask].copy(deep=True)
        logger.debug('{} transmission features found in clipped area with '
                     'minimum max voltage of {}'
                     .format(len(sc_features), tie_line_voltage))

        # Find t-lines connected to substations within clip
        logger.debug('Collecting transmission lines connected to substations')
        mask = sc_features['category'] == SUBSTATION_CAT
        if mask.any():
            trans_gids = sc_features.loc[mask, 'trans_gid'].values
            trans_gids = \
                np.concatenate(self.sub_lines_mapping.loc[trans_gids].values)
            trans_gids = np.unique(trans_gids)
            line_mask = self.features['trans_gid'].isin(trans_gids)
            trans_lines = self.features.loc[line_mask].copy(deep=True)
            line_mask = trans_lines['trans_gid'].isin(sc_features['trans_gid'])
            trans_lines = trans_lines.loc[~line_mask]
            logger.debug('Adding all {} transmission lines connected to '
                         'substations with minimum max voltage of {}'
                         .format(len(trans_lines), tie_line_voltage))
            sc_features = sc_features.append(trans_lines)

        return sc_features, radius

    def process_sc_points(self, capacity_class, sc_point_gids=None, nn_sinks=2,
                          clipping_buffer=1.05, barrier_mult=100,
                          max_workers=None):
        """
        Compute Least Cost Tranmission for desired sc_points

        Parameters
        ----------
        capacity_class : str | int
            Capacity class of transmission features to connect supply curve
            points to
        sc_point_gids : list, optional
            List of sc_point_gids to connect to, by default None
        nn_sinks : int, optional
            Number of nearest neighbor sinks to use for clipping radius
            calculation, by default 2
        clipping_buffer : float, optional
            Buffer to expand clipping radius by, by default 1.05
        barrier_mult : int, optional
            Tranmission barrier multiplier, used when computing the least
            cost tie-line path, by default 100
        max_workers : int, optional
            Number of workers to use for processing, if 1 run in serial,
            if None use all available cores, by default None

        Returns
        -------
        least_costs : pandas.DataFrame
            Least cost connections between all supply curve points and the
            transmission features with the given capacity class that are within
            "nn_sink" nearest infinite sinks
        """
        max_workers = os.cpu_count() if max_workers is None else max_workers

        if sc_point_gids is None:
            sc_point_gids = self.sc_points['sc_point_gid'].values

        tie_line_voltage = self._config.capacity_to_kv(capacity_class)
        least_costs = []
        if max_workers > 1:
            logger.info('Computing Least Cost Transmission for SC points in '
                        'parallel on {} workers'.format(max_workers))
            loggers = [__name__, 'reV', 'reVX']
            with SpawnProcessPool(max_workers=max_workers,
                                  loggers=loggers) as exe:
                futures = []
                for _, sc_point in self.sc_points.iterrows():
                    gid = sc_point['sc_point_gid']
                    if gid in sc_point_gids:
                        sc_features, radius = self._clip_to_sc_point(
                            sc_point, tie_line_voltage, nn_sinks=nn_sinks,
                            clipping_buffer=clipping_buffer)

                        future = exe.submit(TransCapCosts.run,
                                            self._cost_fpath,
                                            sc_point.copy(deep=True),
                                            sc_features, capacity_class,
                                            radius=radius,
                                            xmission_config=self._config,
                                            barrier_mult=barrier_mult,
                                            min_line_length=self._min_line_len)
                        futures.append(future)

                for i, future in enumerate(as_completed(futures)):
                    sc_costs = future.result()
                    if sc_costs is not None:
                        least_costs.append(sc_costs)

                    logger.debug('SC point {} of {} complete!'
                                 .format(i + 1, len(futures)))
                    log_mem(logger)

        else:
            logger.info('Computing Least Cost Transmission for SC points in '
                        'serial')
            i = 1
            for _, sc_point in self.sc_points.iterrows():
                gid = sc_point['sc_point_gid']
                if gid in sc_point_gids:
                    sc_features, radius = self._clip_to_sc_point(
                        sc_point, tie_line_voltage, nn_sinks=nn_sinks,
                        clipping_buffer=clipping_buffer)

                    sc_costs = TransCapCosts.run(
                        self._cost_fpath, sc_point.copy(deep=True),
                        sc_features, capacity_class,
                        radius=radius,
                        xmission_config=self._config,
                        barrier_mult=barrier_mult,
                        min_line_length=self._min_line_len)

                    if sc_costs is not None:
                        least_costs.append(sc_costs)

                    logger.debug('SC point {} of {} complete!'
                                 .format(i, len(sc_point_gids)))
                    log_mem(logger)
                    i += 1

        least_costs = pd.concat(least_costs).sort_values(['sc_point_gid',
                                                          'trans_gid'])
        capacity_class = self._config._parse_cap_class(capacity_class)
        least_costs['max_cap'] = self._config['power_classes'][capacity_class]
        lcp_frac = (len(least_costs['sc_point_gid'].unique())
                    / len(sc_point_gids) * 100)
        logger.info('{:.4f}% of requested sc point gids were succesfully '
                    'mapped to transmission features'.format(lcp_frac))

        return least_costs.reset_index(drop=True)

    @classmethod
    def run(cls, cost_fpath, features_fpath, capacity_class, resolution=128,
            xmission_config=None, sc_point_gids=None, nn_sinks=2,
            clipping_buffer=1.05, barrier_mult=100, max_workers=None):
        """
        Find Least Cost Tranmission connections between desired sc_points to
        given tranmission features for desired capacity class

        Parameters
        ----------
        cost_fpath : str
            Path to h5 file with cost rasters and other required layers
        features_fpath : str
            Path to geopackage with transmission features
        capacity_class : str | int
            Capacity class of transmission features to connect supply curve
            points to
        resolution : int, optional
            SC point resolution, by default 128
        xmission_config : str | dict | XmissionConfig, optional
            Path to Xmission config .json, dictionary of Xmission config
            .jsons, or preloaded XmissionConfig objects, by default None
        sc_point_gids : list, optional
            List of sc_point_gids to connect to, by default None
        nn_sinks : int, optional
            Number of nearest neighbor sinks to use for clipping radius
            calculation, by default 2
        clipping_buffer : float, optional
            Buffer to expand clipping radius by, by default 1.05
        barrier_mult : int, optional
            Tranmission barrier multiplier, used when computing the least
            cost tie-line path, by default 100
        max_workers : int, optional
            Number of workers to use for processing, if 1 run in serial,
            if None use all available cores, by default None

        Returns
        -------
        least_costs : pandas.DataFrame
            Least cost connections between all supply curve points and the
            transmission features with the given capacity class that are within
            "nn_sink" nearest infinite sinks
        """
        ts = time.time()
        lcx = cls(cost_fpath, features_fpath, resolution=resolution,
                  xmission_config=xmission_config)
        least_costs = lcx.process_sc_points(capacity_class,
                                            sc_point_gids=sc_point_gids,
                                            nn_sinks=nn_sinks,
                                            clipping_buffer=clipping_buffer,
                                            barrier_mult=barrier_mult,
                                            max_workers=max_workers)

        logger.info('{} connections were made to {} SC points in {:.4f} '
                    'minutes'
                    .format(len(least_costs),
                            len(least_costs['sc_point_gid'].unique()),
                            (time.time() - ts) / 60))

        return least_costs



# -*- coding: utf-8 -*-
"""
Module to compute least cost xmission paths, distances, and costs one or
more SC points
"""
from concurrent.futures import as_completed
import geopandas as gpd
import logging
import numpy as np
import os
import pandas as pd
from pyproj.crs import CRS
import rasterio
import time
from warnings import warn

from reV.handlers.exclusions import ExclusionLayers
from rex.utilities.execution import SpawnProcessPool
from rex.utilities.loggers import log_mem

from reVX.least_cost_xmission.trans_cap_costs import TieLineCosts
from reVX.utilities.exclusions_converter import ExclusionsConverter

logger = logging.getLogger(__name__)


class LeastCostPaths:
    """
    Compute least cost paths between desired locations
    """
    REQUIRED_LAYRES = ['transmission_barrier']

    def __init__(self, cost_fpath, features_fpath, xmission_config=None):
        """
        Parameters
        ----------
        cost_fpath : str
            Path to h5 file with cost rasters and other required layers
        features_fpath : str
            Path to geopackage with transmission features
        resolution : int, optional
            SC point resolution, by default 128
        xmission_config : str | dict | XmissionConfig, optional
            Path to Xmission config .json, dictionary of Xmission config
            .jsons, or preloaded XmissionConfig objects, by default None
        """
        self._check_layers(cost_fpath)
        self._config = TieLineCosts._parse_config(
            xmission_config=xmission_config)

        self._features, self._row_slice, self._col_slice, self._shape = \
            self._map_to_costs(cost_fpath, gpd.read_file(features_fpath))
        self._features = self._features.drop(columns='geometry')
        self._cost_fpath = cost_fpath

        logger.debug('{} initialized'.format(self))

    def __repr__(self):
        msg = ("{} to be computed for "
               .format(self.__class__.__name__))

        return msg

    @property
    def features(self):
        """
        Table of features to compute paths for

        Returns
        -------
        pandas.DataFrame
        """
        return self._features

    @classmethod
    def _check_layers(cls, cost_fpath):
        """
        Check to make sure the REQUIRED_LAYERS are in cost_fpath

        Parameters
        ----------
        cost_fpath : str
            Path to h5 file with cost rasters and other required layers
        """
        with ExclusionLayers(cost_fpath) as f:
            missing = []
            for lyr in cls.REQUIRED_LAYRES:
                if lyr not in f:
                    missing.append(lyr)

            if missing:
                msg = ("The following layers are required to compute Least "
                       "Cost Transmission but are missing from {}:\n{}"
                       .format(cost_fpath, missing))
                logger.error(msg)
                raise RuntimeError(msg)

    @staticmethod
    def _get_feature_cost_indices(features, crs, transform, shape):
        """
        Map features to cost row, col indicies using rasterio transform

        Parameters
        ----------
        features : gpd.GeoDataFrame
            GeoDataFrame of features to map to cost raster
        crs : pyproj.crs.CRS
            CRS of cost raster
        transform : raster.Affine
            Transform of cost raster
        shape : tuple
            Cost raster shape

        Returns
        -------
        row : ndarray
            Vector of row indicies for each feature
        col : ndarray
            Vector of col indicies for each features
        mask : ndarray
            Boolean mask of features with indicies outside of cost raster
        """
        feat_crs = features.crs.to_dict()
        cost_crs = crs.to_dict()
        bad_crs = ExclusionsConverter._check_crs(cost_crs, feat_crs)
        if bad_crs:
            msg = ('input crs ({}) does not match cost raster crs ({})'
                   ' and will be transformed!'.format(feat_crs, cost_crs))
            logger.warning(msg)
            warn(msg)
            features = features.to_crs(crs)

        logger.debug('Map features to cost raster')
        coords = features['geometry'].centroid
        row, col = rasterio.transform.rowcol(transform, coords.x.values,
                                             coords.y.values)
        row = np.array(row)
        col = np.array(col)

        # Remove features outside of the cost domain
        mask = row >= 0
        mask &= row < shape[0]
        mask &= col >= 0
        mask &= col < shape[1]

        return row, col, mask

    @staticmethod
    def _get_clip_slice(row, col, shape):
        """
        Clip cost raster to bounds of features

        Parameters
        ----------
        row : ndarray
            Vector of row indicies
        col : ndarray
            Vector of col indicies
        shape : tuple
            Full cost array shape

        Returns
        -------
        row_slice : slice
            Row slice to clip too
        col_slice : slice
            Col slice to clip too
        """
        row_slice = slice(max(row.min() - 1, 0), min(row.max() + 1, shape[0]))
        col_slice = slice(max(col.min() - 1, 0), min(col.max() + 1, shape[1]))

        return row_slice, col_slice

    @classmethod
    def _map_to_costs(cls, cost_fpath, features):
        """
        Map features to cost arrays

        Parameters
        ----------
        cost_fpath : str
            Path to h5 file with cost rasters and other required layers
        features : gpd.GeoDataFrame
            GeoDataFrame of features to connect

        Returns
        -------
        features : gpd.GeoDataFrame
            Table of features to compute LeastCostPaths for
        row_slice : slice
            Clipping slice along axis-0 (rows)
        col_slice : slice
            Clipping slice along axis-1 (cols)
        shape : tuple
            Full cost raster shape
        """
        with ExclusionLayers(cost_fpath) as f:
            crs = CRS.from_string(f.crs)
            transform = rasterio.Affine(*f.profile['transform'])
            shape = f.shape

        row, col, mask = cls._get_feature_cost_indices(features, crs,
                                                       transform, shape)
        if any(~mask):
            msg = ("The following features are outside of the cost exclusion "
                   "domain and will be dropped:\n{}"
                   .format(features.loc[~mask]))
            logger.warning(msg)
            row = row[mask]
            col = col[mask]
            features = features.loc[mask].reset_index(drop=True)

        row_slice, col_slice = cls._get_clip_slice(row, col, shape)

        features['row'] = row - row_slice.start
        features['col'] = col - col_slice.start

        return features, row_slice, col_slice, shape

    def _clip_to_feature(self, start_feature_id):
        """
        Clip costs raster to AOI around starting feature to given radius

        Parameters
        ----------
        start_feature_id : int
            Index of start features

        Returns
        -------
        start_idx : ndarray
            (row, col) indicies of start feature
        end_features : pandas.DataFrame
            DataFrame of all end features
        """
        start_idx = \
            self._features.loc[start_feature_id, ['row', 'col']].values
        if len(start_idx.shape) == 2:
            start_idx = start_idx[0]

        end_features = self._features.drop(index=start_feature_id)

        return start_idx, end_features.reset_index(drop=False)

    def process_least_cost_paths(self, capacity_class, barrier_mult=100,
                                 max_workers=None, save_paths=False,):
        """
        Find Least Cost Paths between all pairs of provided features for the
        given tie-line capacity class

        Parameters
        ----------
        capacity_class : str | int
            Capacity class of transmission features to connect supply curve
            points to
        barrier_mult : int, optional
            Tranmission barrier multiplier, used when computing the least
            cost tie-line path, by default 100
        max_workers : int, optional
            Number of workers to use for processing, if 1 run in serial,
            if None use all available cores, by default None
        save_paths : bool, optional
            Flag to save least cost path as a multi-line geometry,
            by default False

        Returns
        -------
        least_cost_paths : pandas.DataFrame | gpd.GeoDataFrame
            DataFrame of lenghts and costs for each path or GeoDataFrame of
            lenght, cost, and geometry for each path
        """
        max_workers = os.cpu_count() if max_workers is None else max_workers

        least_cost_paths = []
        if max_workers > 1:
            logger.info('Computing Least Cost Paths in parallel on {} workers'
                        .format(max_workers))
            log_mem(logger)
            loggers = [__name__, 'reV', 'reVX']
            with SpawnProcessPool(max_workers=max_workers,
                                  loggers=loggers) as exe:
                futures = {}
                for start in self.features.index:
                    start_idx, end_features = \
                        self._clip_to_feature(start_feature_id=start)
                    end_indices = end_features[['row', 'col']].values
                    end_features['start_index'] = start

                    future = exe.submit(TieLineCosts.run,
                                        self._cost_fpath,
                                        start_idx, end_indices,
                                        capacity_class,
                                        self._row_slice, self._col_slice,
                                        barrier_mult=barrier_mult,
                                        save_paths=save_paths)
                    futures[future] = end_features

                for i, future in enumerate(as_completed(futures)):
                    end_features = futures[future]
                    lcp = future.result()
                    lcp = pd.concat((end_features, lcp), axis=1)
                    least_cost_paths.append(lcp)
                    logger.debug('Least cost path {} of {} complete!'
                                 .format(i + 1, len(futures)))
                    log_mem(logger)
        else:
            logger.info('Computing Least Cost Paths in serial')
            log_mem(logger)
            i = 1
            for start in self.features.index:
                start_idx, end_features = \
                    self._clip_to_feature(start_feature_id=start)
                end_indices = end_features[['row', 'col']].values
                end_features = end_features.drop(columns=['row', 'col'])
                end_features['start_index'] = start

                lcp = TieLineCosts.run(self._cost_fpath,
                                       start_idx, end_indices,
                                       capacity_class,
                                       self._row_slice, self._col_slice,
                                       barrier_mult=barrier_mult,
                                       save_paths=save_paths)
                lcp = pd.concat((end_features, lcp), axis=1)
                least_cost_paths.append(lcp)

                logger.debug('Least cost path {} of {} complete!'
                             .format(i, len(self.features)))
                log_mem(logger)
                i += 1

        least_cost_paths = pd.concat(least_cost_paths, ignore_index=True)

        return least_cost_paths

    @classmethod
    def run(cls, cost_fpath, features_fpath, capacity_class,
            xmission_config=None, barrier_mult=100, max_workers=None,
            save_paths=False):
        """
        Find Least Cost Paths between all pairs of provided features for the
        given tie-line capacity class

        Parameters
        ----------
        cost_fpath : str
            Path to h5 file with cost rasters and other required layers
        features_fpath : str
            Path to geopackage with transmission features
        capacity_class : str | int
            Capacity class of transmission features to connect supply curve
            points to
        xmission_config : str | dict | XmissionConfig, optional
            Path to Xmission config .json, dictionary of Xmission config
            .jsons, or preloaded XmissionConfig objects, by default None
        barrier_mult : int, optional
            Tranmission barrier multiplier, used when computing the least
            cost tie-line path, by default 100
        max_workers : int, optional
            Number of workers to use for processing, if 1 run in serial,
            if None use all available cores, by default None
        save_paths : bool, optional
            Flag to save least cost path as a multi-line geometry,
            by default False

        Returns
        -------
        least_cost_paths : pandas.DataFrame | gpd.GeoDataFrame
            DataFrame of lenghts and costs for each path or GeoDataFrame of
            lenght, cost, and geometry for each path
        """
        ts = time.time()
        lcp = cls(cost_fpath, features_fpath, xmission_config=xmission_config)
        least_cost_paths = lcp.process_least_cost_paths(
            capacity_class,
            barrier_mult=barrier_mult,
            save_paths=save_paths,
            max_workers=max_workers)

        logger.info('{} paths were computed in {:.4f} hours'
                    .format(len(least_cost_paths),
                            (time.time() - ts) / 3600))

        return least_cost_paths



# -*- coding: utf-8 -*-
"""
reVX mean wind directions point sub-class
"""
import logging
import numpy as np

from reV.supply_curve.points import AggregationSupplyCurvePoint
from rex.renewable_resource import WindResource

logger = logging.getLogger(__name__)


class MeanWindDirectionsPoint(AggregationSupplyCurvePoint):
    """
    SC point class to compute mean wind directions
    """
    @property
    def h5(self):
        """
        h5 Resource handler object

        Returns
        -------
        _h5 : Resource
            Resource h5 handler object.
        """
        # pylint: disable=E0203,W0201
        if self._h5 is None:
            self._h5 = WindResource(self._h5_fpath)

        return self._h5

    def mean_wind_dirs(self, dset):
        """
        Calc the mean wind directions at every time-step

        Parameters
        ----------
        dset : str
            Wind direction dataset to aggregate

        Returns
        -------
        mean_wind_dirs : np.ndarray | float
            Mean wind direction masked by the binary exclusions
        """
        incl = self.include_mask_flat[self.bool_mask]
        gids = self._gids[self.bool_mask]

        arr_slice = (dset, slice(None), gids)

        angle = np.radians(self.h5[arr_slice], dtype=np.float32)
        sin = np.mean(np.sin(angle) * incl, axis=1)
        cos = np.mean(np.cos(angle) * incl, axis=1)

        mean_wind_dirs = np.degrees(np.arctan2(sin, cos))
        mask = mean_wind_dirs < 0
        mean_wind_dirs[mask] += 360

        return mean_wind_dirs

    @classmethod
    def run(cls, gid, excl, agg_h5, tm_dset, *wind_dir_dsets,
            excl_dict=None, inclusion_mask=None,
            resolution=64, excl_area=0.0081,
            exclusion_shape=None, close=True, gen_index=None):
        """
        Compute exclusions weight mean for the sc point from data

        Parameters
        ----------
        gid : int
            gid for supply curve point to analyze.
        excl : str | ExclusionMask
            Filepath to exclusions h5 or ExclusionMask file handler.
        agg_h5 : str | Resource
            Filepath to .h5 file to aggregate or Resource handler
        tm_dset : str
            Dataset name in the exclusions file containing the
            exclusions-to-resource mapping data.
        wind_dir_dsets : str
            Wind direction dataset(s) to aggregate, can supply multiple
            datasets
        excl_dict : dict | None
            Dictionary of exclusion LayerMask arugments {layer: {kwarg: value}}
            None if excl input is pre-initialized.
        inclusion_mask : np.ndarray
            2D array pre-extracted inclusion mask where 1 is included and 0 is
            excluded. The shape of this will be checked against the input
            resolution.
        resolution : int
            Number of exclusion points per SC point along an axis.
            This number**2 is the total number of exclusion points per
            SC point.
        excl_area : float
            Area of an exclusion cell (square km).
        exclusion_shape : tuple
            Shape of the full exclusions extent (rows, cols). Inputing this
            will speed things up considerably.
        close : bool
            Flag to close object file handlers on exit.
        gen_index : np.ndarray
            Array of generation gids with array index equal to resource gid.
            Array value is -1 if the resource index was not used in the
            generation run.

        Returns
        -------
        out : dict
            Wind direction(s) and meta data aggregated to given supply curve
            point gid
        """
        kwargs = {"excl_dict": excl_dict,
                  "inclusion_mask": inclusion_mask,
                  "resolution": resolution,
                  "excl_area": excl_area,
                  "exclusion_shape": exclusion_shape,
                  "close": close,
                  "gen_index": gen_index}

        with cls(gid, excl, agg_h5, tm_dset, **kwargs) as point:
            out = {'meta': point.summary}
            for dset in wind_dir_dsets:
                out[dset] = point.mean_wind_dirs(dset)

        return out



# -*- coding: utf-8 -*-
"""
reVX Least Cost Xmission Configurations
"""

from reV.config.base_analysis_config import AnalysisConfig


class CostCreatorConfig(AnalysisConfig):
    """Config framework for creating cost layers"""

    NAME = 'CostCreator'
    REQUIREMENTS = ('h5_fpath', 'iso_regions')

    def __init__(self, config):
        """
        Parameters
        ----------
        config : str | dict
            Path to config .json or pre-extracted config input dictionary.
        """
        super().__init__(config)
        self._default_slope_layer = 'srtm_slope'
        self._default_nlcd_layer = 'usa_mrlc_nlcd2011'

    @property
    def h5_fpath(self):
        """
        H5 file to save costs to (required).
        """
        return self['h5_fpath']

    @property
    def iso_regions(self):
        """
        File with raster of ISO regions
        """
        return self['iso_regions']

    @property
    def excl_h5(self):
        """
        Path to exclusion .h5 file containing NLCD and
        slope layers, if None use h5_fpath if None assume
        NLCD and slope layers are in self._excl_h5
        """
        return self.get('excl_h5', None)

    @property
    def cost_configs(self):
        """
        JSON file with cost configs
        """
        return self.get('cost_configs', None)

    @property
    def slope_layer(self):
        """
        Name of slope layer in excl_h5
        """
        return self.get('slope_layer', self._default_slope_layer)

    @property
    def nlcd_layer(self):
        """
        Name of NLCD (land use) layer in excl_h5
        """
        return self.get('nlcd_layer', self._default_nlcd_layer)

    @property
    def default_mults(self):
        """
        JSON of Multipliers for regions not specified in
        iso_mults_fpath
        """
        return self.get('default_mults', None)

    @property
    def tiff_dir(self):
        """
        JSON with Extra layers to add to h5 file, for example dist_to_coast
        """
        return self.get('tiff_dir', None)

    @property
    def extra_layers(self):
        """
        Path to save costs and intermediary rasters as geotiffs
        """
        return self.get('extra_layers', None)


class LeastCostXmissionConfig(AnalysisConfig):
    """Config framework for Least Cost Xmission"""

    NAME = 'LeastCostXmission'
    REQUIREMENTS = ('cost_fpath', 'features_fpath', 'capacity_class')

    def __init__(self, config):
        """
        Parameters
        ----------
        config : str | dict
            Path to config .json or pre-extracted config input dictionary.
        """
        super().__init__(config)
        self._default_resolution = 128
        self._default_nn_sinks = 2
        self._default_clipping_buffer = 1.05
        self._default_barrier_mult = 100
        self._default_min_line_length = 5.7

    @property
    def cost_fpath(self):
        """
        .h5 file with cost and required layers
        """
        return self['cost_fpath']

    @property
    def features_fpath(self):
        """
        Tranmission feature .gpkg
        """
        return self['features_fpath']

    @property
    def capacity_class(self):
        """
        Capacity class, either {capacity}MW or capacity value in MW
        """
        return self['capacity_class']

    @property
    def resolution(self):
        """
        SC point resolution
        """
        return self.get('resolution', self._default_resolution)

    @property
    def xmission_config(self):
        """
        Xmission config input
        """
        return self.get('xmission_config', None)

    @property
    def sc_point_gids(self):
        """
        List of sc_point_gids to compute Least Cost Xmission for
        """
        return self.get('sc_point_gids', None)

    @property
    def nn_sinks(self):
        """
        Number of nearest neighbor sinks to use for clipping radius
        """
        return self.get('nn_sinks', self._default_nn_sinks)

    @property
    def clipping_buffer(self):
        """
        Buffer to add to clipping radius
        """
        return self.get('clipping_buffer', self._default_clipping_buffer)

    @property
    def barrier_mult(self):
        """
        Transmission barrier multiplier to use for MCP costs
        """
        return self.get('barrier_mult', self._default_barrier_mult)


class LeastCostPathsConfig(AnalysisConfig):
    """Config framework for Least Cost Paths"""

    NAME = 'LeastCostPaths'
    REQUIREMENTS = ('cost_fpath', 'features_fpath', 'capacity_class')

    def __init__(self, config):
        """
        Parameters
        ----------
        config : str | dict
            Path to config .json or pre-extracted config input dictionary.
        """
        super().__init__(config)
        self._default_barrier_mult = 100

    @property
    def cost_fpath(self):
        """
        .h5 file with cost and required layers
        """
        return self['cost_fpath']

    @property
    def features_fpath(self):
        """
        Tranmission feature .gpkg
        """
        return self['features_fpath']

    @property
    def capacity_class(self):
        """
        Capacity class, either {capacity}MW or capacity value in MW
        """
        return self['capacity_class']

    @property
    def xmission_config(self):
        """
        Xmission config input
        """
        return self.get('xmission_config', None)

    @property
    def barrier_mult(self):
        """
        Transmission barrier multiplier to use for MCP costs
        """
        return self.get('barrier_mult', self._default_barrier_mult)

    @property
    def save_paths(self):
        """
        Boolean flag to save the least cost paths along with the costs and
        lengths
        """
        return self.get('save_paths', False)



# -*- coding: utf-8 -*-
"""
Turbine Flicker exclusions calculator
"""
from concurrent.futures import as_completed
import logging
import numpy as np
import os
from warnings import warn

from reV.handlers.exclusions import ExclusionLayers
from reV.supply_curve.extent import SupplyCurveExtent
from reV.supply_curve.tech_mapping import TechMapping
from reVX.wind_dirs.mean_wind_dirs_point import MeanWindDirectionsPoint
from reVX.utilities.exclusions_converter import ExclusionsConverter
from rex.resource_extraction.resource_extraction import WindX
from rex.utilities.execution import SpawnProcessPool
from rex.utilities.loggers import log_mem

logger = logging.getLogger(__name__)


class TurbineFlicker:
    """
    Class to compute turbine shadow flicker and exclude sites that will
    cause excessive flicker on building
    """
    STEPS_PER_HOUR = 1
    GRIDCELL_SIZE = 90
    FLICKER_ARRAY_LEN = 65

    def __init__(self, excl_fpath, res_fpath, building_layer,
                 resolution=640, tm_dset='techmap_wtk'):
        """
        Parameters
        ----------
        excl_fpath : str
            Filepath to exclusions h5 file. File must contain "building_layer"
            and "tm_dset".
        res_fpath : str
            Filepath to wind resource .h5 file containing hourly wind
            direction data
        building_layer : str
            Exclusion layer containing buildings from which turbine flicker
            exclusions will be computed.
        tm_dset : str, optional
            Dataset / layer name for wind toolkit techmap,
            by default 'techmap_wtk'
        resolution : int, optional
            SC resolution, must be input in combination with gid,
            by default 640
        """
        self._excl_h5 = excl_fpath
        self._res_h5 = res_fpath
        self._bld_layer = building_layer
        self._res = resolution
        self._preflight_check(tm_dset=tm_dset)
        self._sc_points = self._get_sc_points(tm_dset=tm_dset)

    def __repr__(self):
        msg = ("{} from {}"
               .format(self.__class__.__name__, self._bld_layer))

        return msg

    @classmethod
    def _compute_shadow_flicker(cls, lat, lon, blade_length, wind_dir):
        """
        Compute shadow flicker for given location

        Parameters
        ----------
        lat : float
            Latitude coordinate of turbine
        lon : float
            Longitude coordinate of turbine
        blade_length : float
            Turbine blade length. Hub height = 2.5 * blade length
        wind_dir : ndarray
            Time-series of wind direction for turbine

        Returns
        -------
        shadow_flicker : ndarray
            2D array centered on the turbine with the number of flicker hours
            per "exclusion" pixel
        """
        # Import HOPP dynamically so its not a requirement
        from hybrid.flicker.flicker_mismatch_grid import FlickerMismatch

        mult = (cls.FLICKER_ARRAY_LEN * cls.GRIDCELL_SIZE) / 2
        mult = mult / (blade_length * 2)
        FlickerMismatch.diam_mult_nwe = mult
        FlickerMismatch.diam_mult_s = mult
        FlickerMismatch.steps_per_hour = cls.STEPS_PER_HOUR
        FlickerMismatch.turbine_tower_shadow = False

        assert len(wind_dir) == 8760

        shadow_flicker = FlickerMismatch(lat, lon,
                                         blade_length=blade_length,
                                         angles_per_step=None,
                                         wind_dir=wind_dir,
                                         gridcell_height=cls.GRIDCELL_SIZE,
                                         gridcell_width=cls.GRIDCELL_SIZE,
                                         gridcells_per_string=1)
        shadow_flicker = shadow_flicker.create_heat_maps(range(0, 8760),
                                                         ("time", ))[0]

        return shadow_flicker

    @staticmethod
    def _invert_shadow_flicker_arr(shadow_flicker):
        """
        Check to ensure the shadow_flicker array is odd in shape, i.e. both
        dimensions are odd allowing for a central pixel for the turbine to
        sit on. Flip both axes to mimic the turbine sitting on each building.
        All flicker pixels will now indicate locations where a turbine would
        need to be to cause flicker on said building

        Parameters
        ----------
        shadow_flicker : ndarray
            2D array centered on the turbine with the number of flicker hours
            per "exclusion" pixel

        Returns
        -------
        shadow_flicker : ndarray
            Inverted 2D shadow flicker array with odd dimensions if needed.
        """
        reduce_slice = ()
        reduce_arr = False
        for s in shadow_flicker.shape:
            if s % 2:
                reduce_slice += (slice(None), )
            else:
                reduce_slice += (slice(0, -1), )
                reduce_arr = True

        if reduce_arr:
            shape_in = shadow_flicker.shape
            shadow_flicker = shadow_flicker[reduce_slice]
            msg = ('Shadow flicker array with shape {} does not have a '
                   'central pixel! Shade has been reduced to {}!'
                   .format(shape_in, shadow_flicker.shape))
            logger.warning(msg)
            warn(msg)

        return shadow_flicker[::-1, ::-1]

    @classmethod
    def _get_flicker_excl_shifts(cls, shadow_flicker, flicker_threshold=30):
        """
        Determine locations of shadow flicker that exceed the given threshold,
        convert to row and column shifts. These are the locations turbines
        would need to in relation to building to cause flicker exceeding the
        threshold value.

        Parameters
        ----------
        shadow_flicker : ndarray
            2D array centered on the turbine with the number of flicker hours
            per "exclusion" pixel
        flicker_threshold : int, optional
            Maximum number of allowable flicker hours, by default 30

        Returns
        -------
        row_shifts : ndarray
            Shifts along axis 0 from building location to pixels to be excluded
        col_shifts : ndarray
            Shifts along axis 1 from building location to pixels to be excluded
        """
        # ensure shadow_flicker array is regularly shaped and invert for
        # mapping to buildings
        shadow_flicker = cls._invert_shadow_flicker_arr(shadow_flicker)

        # normalize by number of time-steps to match shadow flicker results
        flicker_threshold /= 8760
        shape = shadow_flicker.shape
        row_shifts, col_shifts = np.where(shadow_flicker > flicker_threshold)
        check = (np.any(np.isin(row_shifts, [0, shape[0] - 1]))
                 or np.any(np.isin(col_shifts, [0, shape[1] - 1])))
        if check:
            msg = ("Turbine flicker exceeding {} appears to extend beyond the "
                   "FlickerModel domain! Please increase the "
                   "FLICKER_ARRAY_LEN and try again!")
            logger.error(msg)
            raise RuntimeError(msg)

        row_shifts -= shape[0] // 2
        col_shifts -= shape[1] // 2

        return row_shifts, col_shifts

    @staticmethod
    def _get_building_indices(excl_fpath, building_layer, gid,
                              resolution=640, building_threshold=0):
        """
        Find buildings in sc point sub-array and convert indices to full
        exclusion indices

        Parameters
        ----------
        excl_fpath : str
            Filepath to exclusions h5 file. File must contain "building_layer"
            and "tm_dset".
        building_layer : str
            Exclusion layer containing buildings from which turbine flicker
            exclusions will be computed.
        gid : int
            sc point gid to extract buildings for
        resolution : int, optional
            SC resolution, must be input in combination with gid,
            by default 640
        building_threshold : float, optional
            Threshold for exclusion layer values to identify pixels with
            buildings, values are % of pixel containing a building,
            by default 0

        Returns
        -------
        row_idx : ndarray
            Axis 0 indices of building in sc point sub-array in full exclusion
            array
        col_idx : ndarray
            Axis 1 indices of building in sc point sub-array in full exclusion
            array
        shape : tuple
            Full exclusion array shape
        """
        with ExclusionLayers(excl_fpath) as f:
            shape = f.shape
            row_slice, col_slice = MeanWindDirectionsPoint.get_agg_slices(
                gid, shape, resolution)

            sc_blds = f[building_layer, row_slice, col_slice]

        row_idx = np.array(range(*row_slice.indices(row_slice.stop)))
        col_idx = np.array(range(*col_slice.indices(col_slice.stop)))
        bld_row_idx, bld_col_idx = np.where(sc_blds > building_threshold)

        return row_idx[bld_row_idx], col_idx[bld_col_idx], shape

    @staticmethod
    def _create_excl_indices(bld_idx, flicker_shifts, shape):
        """
        Create 2D (row, col) indices of pixels to be excluded based on
        building indices and shadow flicker shifts.

        Parameters
        ----------
        bld_idx : tuple
            (row, col) indices of building onto which shadow flicker exclusions
            are to be mapped.
        flicker_shifts : tuple
            Index shifts (row, col) from building locations to exclude based
            on shadow flicker results. Shifts are based on shadow flicker
            threshold. Shadow flicker array is inverted to represent mapping of
            shadow onto buildings
        shape : tuple
            Full exclusion array shape

        Returns
        -------
        excl_row_idx : ndarray
            Row (axis 0) indices of pixels to be excluded because they will
            cause excessive shadow flicker on building in supply curve point
            gid subset
        excl_col_idx : ndarray
            Column (axis 1) indices of pixels to be excluded because they will
            cause excessive shadow flicker on building in supply curve point
            gid subset
        """
        row_idx, col_idx = bld_idx
        row_shifts, col_shifts = flicker_shifts

        excl_row_idx = (row_idx + row_shifts[:, None]).ravel()
        excl_row_idx[excl_row_idx < 0] = 0
        excl_row_idx[excl_row_idx >= shape[0]] = shape[0] - 1

        excl_col_idx = (col_idx + col_shifts[:, None]).ravel()
        excl_col_idx[excl_col_idx < 0] = 0
        excl_col_idx[excl_col_idx >= shape[1]] = shape[1] - 1

        return excl_row_idx.astype(np.uint32), excl_col_idx.astype(np.uint32)

    @classmethod
    def _exclude_turbine_flicker(cls, point, excl_fpath, res_fpath,
                                 building_layer, hub_height,
                                 building_threshold=0, flicker_threshold=30,
                                 resolution=640):
        """
        Exclude all pixels that will cause flicker exceeding the
        "flicker_threshold" on buildings that exist within
        supply curve point gid subset of "building_layer". Buildings
        are defined as pixels with >= the "building_threshold".
        Shadow flicker is computed at the supply curve point
        resolution and applied to all buildings within that supply curve point
        sub-array. Excluded pixels can extend beyond the supply curve point
        gid subset, for example if a building sits at the edge of the subset.

        Parameters
        ----------
        gid : int
            Supply curve point gid to aggregate wind directions for
        excl_fpath : str
            Filepath to exclusions h5 file. File must contain "tm_dset".
        res_fpath : str
            Filepath to wind resource .h5 file containing hourly wind
            direction data
        building_layer : str
            Exclusion layer containing buildings from which turbine flicker
            exclusions will be computed.
        hub_height : int
            Hub-height in meters to compute turbine shadow flicker for
        building_threshold : float, optional
            Threshold for exclusion layer values to identify pixels with
            buildings, values are % of pixel containing a building,
            by default 0
        flicker_threshold : int, optional
            Maximum number of allowable flicker hours, by default 30
        resolution : int, optional
            SC resolution, must be input in combination with gid,
            by default 640

        Returns
        -------
        excl_idx : tuple
            (row, col) indices of pixels to be excluded because they will cause
            excessive shadow flicker on building in supply curve point gid
            subset
        """
        row_idx, col_idx, shape = cls._get_building_indices(
            excl_fpath, building_layer, point.name,
            resolution=resolution, building_threshold=building_threshold)

        with WindX(res_fpath, log_vers=False) as f:
            dset = 'winddirection_{}m'.format(hub_height)
            wind_dir = f[dset, :, int(point['res_gid'])]

            if len(wind_dir) == 8784:
                wind_dir = wind_dir[:-24]

        blade_length = hub_height / 2.5
        shadow_flicker = cls._compute_shadow_flicker(point['latitude'],
                                                     point['longitude'],
                                                     blade_length,
                                                     wind_dir)

        flicker_shifts = cls._get_flicker_excl_shifts(
            shadow_flicker, flicker_threshold=flicker_threshold)

        excl_idx = cls._create_excl_indices((row_idx, col_idx),
                                            flicker_shifts, shape)

        return excl_idx

    def _preflight_check(self, tm_dset='techmap_wtk'):
        """
        Check to ensure building_layer and tm_dset are in exclusion .h5 file

        Parameters
        ----------
        tm_dset : str, optional
            Dataset / layer name for wind toolkit techmap,
            by default 'techmap_wtk'
        """
        with ExclusionLayers(self._excl_h5) as f:
            layers = f.layers

        if self._bld_layer not in layers:
            msg = ("{} is not available in {}"
                   .format(self._bld_layer, self._excl_h5))
            logger.error(msg)
            raise RuntimeError(msg)

        if tm_dset not in layers:
            logger.warning('Could not find techmap "{t}" in {e}. '
                           'Creating {t} using reV TechMapping'
                           .format(t=tm_dset, e=self._excl_h5))
            try:
                TechMapping.run(self._excl_h5, self._res_h5,
                                dset=tm_dset)
            except Exception as e:
                logger.exception('TechMapping process failed. Received the '
                                 'following error:\n{}'.format(e))
                raise e

    def _get_sc_points(self, tm_dset='techmap_wtk'):
        """
        Get the valid sc points to run turbine flicker for

        Parameters
        ----------
        tm_dset : str, optional
            [description], by default 'techmap_wtk'

        Returns
        -------
        points : pandas.DataFrame
            DataFrame of valid sc point gids with their latitude and longitude
            coordinates and nearest resource gid
        """
        with SupplyCurveExtent(self._excl_h5, resolution=self._res) as sc:
            points = sc.points
            points['latitude'] = sc.latitude
            points['longitude'] = sc.longitude
            gids = sc.valid_sc_points(tm_dset)
            points = points.loc[gids]

        with WindX(self._res_h5) as f:
            res_gids = f.lat_lon_gid(points[['latitude', 'longitude']].values,
                                     check_lat_lon=False)
            points['res_gid'] = res_gids

        return points

    def compute_exclusions(self, hub_height, building_threshold=0,
                           flicker_threshold=30, max_workers=None,
                           out_layer=None):
        """
        Exclude all pixels that will cause flicker exceeding the
        "flicker_threshold" on any building in "building_layer". Buildings
        are defined as pixels with >= the "building_threshold value in
        "building_layer". Shadow flicker is computed at the supply curve point
        resolution based on a turbine with "hub_height" (m) and applied to all
        buildings within that supply curve point sub-array.

        Parameters
        ----------
        hub_height : int
            Hub-height in meters to compute turbine shadow flicker for
        building_threshold : float, optional
            Threshold for exclusion layer values to identify pixels with
            buildings, values are % of pixel containing a building,
            by default 0
        flicker_threshold : int, optional
            Maximum number of allowable flicker hours, by default 30
        resolution : int, optional
            SC resolution, must be input in combination with gid,
            by default 640
        max_workers : None | int, optional
            Number of workers to use, if 1 run in serial, if None use all
            available cores, by default None
        out_layer : str, optional
            Layer to save exclusions under. Layer will be saved in
            "excl_fpath", by default None

        Returns
        -------
        flicker_arr : ndarray
            2D inclusion array. Pixels to exclude (0) to prevent shadow
            flicker on buildings in "building_layer"
        """
        with ExclusionLayers(self._excl_h5) as f:
            exclusion_shape = f.shape
            profile = f.profile

        if max_workers is None:
            max_workers = os.cpu_count()

        etf_kwargs = {"building_threshold": building_threshold,
                      "flicker_threshold": flicker_threshold,
                      "resolution": self._res}
        flicker_arr = np.ones(exclusion_shape, dtype=np.uint8)
        if max_workers > 1:
            msg = ('Computing exclusions from {} based on {}m turbines '
                   'in parallel using {} workers'
                   .format(self, hub_height, max_workers))
            logger.info(msg)

            loggers = [__name__, 'reVX', 'rex']
            with SpawnProcessPool(max_workers=max_workers,
                                  loggers=loggers) as exe:
                futures = []
                for _, point in self._sc_points.iterrows():
                    future = exe.submit(self._exclude_turbine_flicker,
                                        point, self._excl_h5, self._res_h5,
                                        self._bld_layer, hub_height,
                                        **etf_kwargs)
                    futures.append(future)

                for i, future in enumerate(as_completed(futures)):
                    row_idx, col_idx = future.result()
                    flicker_arr[row_idx, col_idx] = 0
                    logger.debug('Completed {} out of {} gids'
                                 .format((i + 1), len(futures)))
                    log_mem(logger)
        else:
            msg = ('Computing exclusions from {} based on {}m turbines in '
                   'serial'.format(self, hub_height))
            logger.info(msg)
            for i, (_, point) in enumerate(self._sc_points.iterrows()):
                row_idx, col_idx = self._exclude_turbine_flicker(
                    point, self._excl_h5, self._res_h5, self._bld_layer,
                    hub_height, **etf_kwargs)
                flicker_arr[row_idx, col_idx] = 0
                logger.debug('Completed {} out of {} gids'
                             .format((i + 1), len(self._sc_points)))
                log_mem(logger)

        if out_layer:
            logger.info('Saving flicker exclusions to {} as {}'
                        .format(self._excl_h5, out_layer))
            description = ("Pixels with value 1 will cause greater than {} "
                           "hours of flicker on buildings in {}. Shadow "
                           "flicker is computed using a {}m turbine."
                           .format(flicker_threshold, self._bld_layer,
                                   hub_height))
            ExclusionsConverter._write_layer(self._excl_h5, out_layer,
                                             profile, flicker_arr,
                                             description=description)

        return flicker_arr

    @classmethod
    def run(cls, excl_fpath, res_fpath, building_layer, hub_height,
            tm_dset='techmap_wtk', building_threshold=0,
            flicker_threshold=30, resolution=640,
            max_workers=None, out_layer=None):
        """
        Exclude all pixels that will cause flicker exceeding the
        "flicker_threshold" on any building in "building_layer". Buildings
        are defined as pixels with >= the "building_threshold value in
        "building_layer". Shadow flicker is computed at the supply curve point
        resolution based on a turbine with "hub_height" (m) and applied to all
        buildings within that supply curve point sub-array.

        Parameters
        ----------
        excl_fpath : str
            Filepath to exclusions h5 file. File must contain "building_layer"
            and "tm_dset".
        res_fpath : str
            Filepath to wind resource .h5 file containing hourly wind
            direction data
        building_layer : str
            Exclusion layer containing buildings from which turbine flicker
            exclusions will be computed.
        hub_height : int
            Hub-height in meters to compute turbine shadow flicker for
        tm_dset : str, optional
            Dataset / layer name for wind toolkit techmap,
            by default 'techmap_wtk'
        building_threshold : float, optional
            Threshold for exclusion layer values to identify pixels with
            buildings, values are % of pixel containing a building,
            by default 0
        flicker_threshold : int, optional
            Maximum number of allowable flicker hours, by default 30
        resolution : int, optional
            SC resolution, must be input in combination with gid,
            by default 640
        max_workers : None | int, optional
            Number of workers to use, if 1 run in serial, if None use all
            available cores, by default None
        out_layer : str, optional
            Layer to save exclusions under. Layer will be saved in
            "excl_fpath", by default None

        Returns
        -------
        flicker_arr : ndarray
            2D inclusion array. Pixels to exclude (0) to prevent shadow
            flicker on buildings in "building_layer"
        """
        flicker = cls(excl_fpath, res_fpath, building_layer,
                      resolution=resolution,
                      tm_dset=tm_dset)
        out_excl = flicker.compute_exclusions(
            hub_height,
            building_threshold=building_threshold,
            flicker_threshold=flicker_threshold,
            max_workers=max_workers,
            out_layer=out_layer)

        return out_excl



# -*- coding: utf-8 -*-
# pylint: disable=all
"""
Xmission Cost Creator Command Line Interface
"""
import click
import logging
import os

from rex.utilities.loggers import init_mult
from rex.utilities.cli_dtypes import STR
from rex.utilities.hpc import SLURM
from rex.utilities.utilities import get_class_properties, safe_json_load

from reVX.config.least_cost_xmission import CostCreatorConfig
from reVX.least_cost_xmission.cost_creator import XmissionCostCreator
from reVX import __version__

logger = logging.getLogger(__name__)


@click.group()
@click.version_option(version=__version__)
@click.option('--name', '-n', default='CostCreator', type=STR,
              show_default=True,
              help='Job name.')
@click.option('--verbose', '-v', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def main(ctx, name, verbose):
    """
    Least Cost Xmission Cost Creator Command Line Interface
    """
    ctx.ensure_object(dict)
    ctx.obj['VERBOSE'] = verbose
    ctx.obj['NAME'] = name


@main.command()
def valid_config_keys():
    """
    Echo the valid Least Cost Xmission Cost Creator config keys
    """
    click.echo(', '.join(get_class_properties(CostCreatorConfig)))


def run_local(ctx, config):
    """
    Run Least Cost Xmission Cost Creator locally using config

    Parameters
    ----------
    ctx : click.ctx
        click ctx object
    config : reVX.config.least_cost_xmission.CostCreatorConfig
        Cost Creator config object.
    """
    ctx.obj['NAME'] = config.name
    ctx.invoke(local,
               h5_fpath=config.h5_fpath,
               iso_regions=config.iso_regions,
               excl_h5=config.excl_h5,
               cost_configs=config.cost_configs,
               slope_layer=config.slope_layer,
               nlcd_layer=config.nlcd_layer,
               default_mults=config.default_mults,
               tiff_dir=config.tiff_dir,
               extra_layers=config.extra_layers,
               log_dir=config.log_directory,
               verbose=config.log_level)


@main.command()
@click.option('--config', '-c', required=True,
              type=click.Path(exists=True),
              help='Filepath to AssemblyAreas config json file.')
@click.option('--verbose', '-v', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def from_config(ctx, config, verbose):
    """
    Run Least Cost Xmission Cost Creator from a config.
    """

    config = CostCreatorConfig(config)

    if 'VERBOSE' in ctx.obj:
        if any((ctx.obj['VERBOSE'], verbose)):
            config._log_level = logging.DEBUG
    elif verbose:
        config._log_level = logging.DEBUG

    if config.execution_control.option == 'local':
        run_local(ctx, config)

    if config.execution_control.option == 'eagle':
        eagle(config)


@main.command()
@click.option('--h5_fpath', '-h5', type=click.Path(),
              required=True,
              help=("H5 file to save costs to"))
@click.option('--iso_regions', '-iso', required=True,
              type=click.Path(exists=True),
              help="File with raster of ISO regions.")
@click.option('--excl_h5', '-excl', type=STR,
              show_default=True, default=None,
              help=("Path to exclusion .h5 file containing NLCD and "
                    "slope layers, if None use h5_fpath if None assume "
                    "NLCD and slope layers are in self._excl_h5"))
@click.option('--cost_configs', '-ccfg', type=STR,
              show_default=True, default=None,
              help=("JSON file with cost configs"))
@click.option('--slope_layer', '-slope', type=str, show_default=True,
              default="srtm_slope",
              help=("Name of slope layer in excl_h5"))
@click.option('--nlcd_layer', '-nlcd', type=str, show_default=True,
              default="usa_mrlc_nlcd2011",
              help=("Name of NLCD (land use) layer in excl_h5"))
@click.option('--default_mults', '-dm', type=STR,
              show_default=True, default=None,
              help=("JSON of Multipliers for regions not specified in "
                    "iso_mults_fpath"))
@click.option('--tiff_dir', '-tiff', type=STR,
              show_default=True, default=None,
              help=("Path to save costs and intermediary rasters as geotiffs"))
@click.option('--extra_layers', '-lyrs', type=STR,
              show_default=True, default=None,
              help=("JSON with Extra layers to add to h5 file, for example "
                    "dist_to_coast"))
@click.option('--log_dir', '-log', default=None, type=STR,
              show_default=True,
              help='Directory to dump log files.')
@click.option('--verbose', '-v', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def local(ctx, h5_fpath, iso_regions, excl_h5, cost_configs, slope_layer,
          nlcd_layer, default_mults, tiff_dir, extra_layers, log_dir, verbose):
    """
    Run Least Cost Xmission Cost Creator on local hardware
    """
    name = ctx.obj['NAME']
    if 'VERBOSE' in ctx.obj:
        verbose = any((ctx.obj['VERBOSE'], verbose))

    log_modules = [__name__, 'reVX', 'reV', 'rex']
    init_mult(name, log_dir, modules=log_modules, verbose=verbose)

    if isinstance(default_mults, str):
        default_mults = safe_json_load(default_mults)

    if isinstance(extra_layers, str):
        extra_layers = safe_json_load(extra_layers)

    logger.info('Computing Xmission Cost layers and writing them to {}'
                .format(h5_fpath))
    XmissionCostCreator.run(h5_fpath, iso_regions, excl_h5=excl_h5,
                            cost_configs=cost_configs,
                            slope_layer=slope_layer, nlcd_layer=nlcd_layer,
                            default_mults=default_mults, tiff_dir=tiff_dir,
                            extra_layers=extra_layers)


def get_node_cmd(config):
    """
    Get the node CLI call for Least Cost Xmission Cost Creator

    Parameters
    ----------
    config : reVX.config.least_cost_xmission.CostCreatorConfig
        Cost Creator config object.

    Returns
    -------
    cmd : str
        CLI call to submit to SLURM execution.
    """

    args = ['-n {}'.format(SLURM.s(config.name)),
            'local',
            '-h5 {}'.format(SLURM.s(config.h5_fpath)),
            '-iso {}'.format(SLURM.s(config.iso_regions)),
            '-excl {}'.format(SLURM.s(config.excl_h5)),
            '-ccfg {}'.format(SLURM.s(config.cost_configs)),
            '-slope {}'.format(SLURM.s(config.slope_layer)),
            '-nlcd {}'.format(SLURM.s(config.nlcd_layer)),
            '-dm {}'.format(SLURM.s(config.default_mults)),
            '-tiff {}'.format(SLURM.s(config.tiff_dir)),
            '-lyrs {}'.format(SLURM.s(config.extra_layers)),
            '-log {}'.format(SLURM.s(config.log_directory)),
            ]

    if config.log_level == logging.DEBUG:
        args.append('-v')

    cmd = ('python -m reVX.least_cost_xmission.cost_creator_cli {}'
           .format(' '.join(args)))
    logger.debug('Submitting the following cli call:\n\t{}'.format(cmd))

    return cmd


def eagle(config):
    """
    Run Least Cost Xmission Cost Creator on Eagle HPC.

    Parameters
    ----------
    config : reVX.config.least_cost_xmission.CostCreatorConfig
        Cost Creator config object.
    """

    cmd = get_node_cmd(config)
    name = config.name
    log_dir = config.log_directory
    stdout_path = os.path.join(log_dir, 'stdout/')

    slurm_manager = SLURM()

    logger.info('Running Least Cost Xmission Cost Creator on Eagle with '
                'node name "{}"'.format(name))
    out = slurm_manager.sbatch(cmd,
                               alloc=config.execution_control.allocation,
                               memory=config.execution_control.memory,
                               walltime=config.execution_control.walltime,
                               feature=config.execution_control.feature,
                               name=name, stdout_path=stdout_path,
                               conda_env=config.execution_control.conda_env,
                               module=config.execution_control.module)[0]
    if out:
        msg = ('Kicked off Least Cost Xmission Cost Creator "{}" '
               '(SLURM jobid #{}) on Eagle.'
               .format(name, out))
    else:
        msg = ('Was unable to kick off Least Cost Xmission Cost Creator '
               '"{}". Please see the stdout error messages'
               .format(name))

    click.echo(msg)
    logger.info(msg)


if __name__ == '__main__':
    try:
        main(obj={})
    except Exception:
        logger.exception('Error running Least Cost Xmission Cost Creator CLI')
        raise



# -*- coding: utf-8 -*-
"""
Compute the mean wind direction for each supply curve point
"""
import logging
import numpy as np

from reV.supply_curve.aggregation import Aggregation, AggFileHandler
from reV.supply_curve.extent import SupplyCurveExtent
from reV.utilities.exceptions import EmptySupplyCurvePointError
from reVX.wind_dirs.mean_wind_dirs_point import MeanWindDirectionsPoint
from reVX.utilities.utilities import log_versions
from rex.utilities.loggers import log_mem

logger = logging.getLogger(__name__)


class MeanWindDirections(Aggregation):
    """
    Average the wind direction via the wind vectors.
    Then convert to equivalent sc_point_gid
    """

    def __init__(self, res_h5_fpath, excl_fpath, wdir_dsets,
                 tm_dset='techmap_wtk', excl_dict=None,
                 area_filter_kernel='queen', min_area=None,
                 resolution=128, excl_area=None):
        """
        Parameters
        ----------
        res_h5_fpath : str
            Filepath to .h5 file containing wind direction data
        excl_fpath : str
            Filepath to exclusions h5 with techmap dataset.
        wdir_dsets : str | list
            Wind direction dataset to average
        tm_dset : str, optional
            Dataset name in the techmap file containing the
            exclusions-to-resource mapping data,
            by default 'techmap_wtk'
        excl_dict : dict | None, optional
            Dictionary of exclusion LayerMask arugments {layer: {kwarg: value}}
            by default None
        area_filter_kernel : str, optional
            Contiguous area filter method to use on final exclusions mask,
            by default 'queen'
        min_area : float | None, optional
            Minimum required contiguous area filter in sq-km, by default None
        resolution : int | None, optional
            SC resolution, must be input in combination with gid,
            by default 128
        excl_area : float | None, optional
            Area of an exclusion pixel in km2. None will try to infer the area
            from the profile transform attribute in excl_fpath,
            by default None
        """
        log_versions(logger)
        if isinstance(wdir_dsets, str):
            wdir_dsets = [wdir_dsets]

        for dset in wdir_dsets:
            if not dset.startswith('winddirection'):
                msg = ('{} is not a valid wind direction dataset!'
                       .format(dset))
                logger.error(msg)
                raise ValueError(msg)

        super().__init__(excl_fpath, res_h5_fpath, tm_dset, *wdir_dsets,
                         excl_dict=excl_dict,
                         area_filter_kernel=area_filter_kernel,
                         min_area=min_area,
                         resolution=resolution, excl_area=excl_area)

    # pylint: disable=unused-argument
    @classmethod
    def run_serial(cls, excl_fpath, h5_fpath, tm_dset, *wind_dir_dset,
                   excl_dict=None, inclusion_mask=None,
                   area_filter_kernel='queen', min_area=None,
                   resolution=128, excl_area=0.0081, gids=None,
                   gen_index=None, **kwargs):
        """
        Standalone method to aggregate - can be parallelized.

        Parameters
        ----------
        excl_fpath : str | list | tuple
            Filepath to exclusions h5 with techmap dataset
            (can be one or more filepaths).
        h5_fpath : str
            Filepath to .h5 file to aggregate
        tm_dset : str
            Dataset name in the techmap file containing the
            exclusions-to-resource mapping data.
        wind_dir_dset : str
            Wind directions to aggreate, can supply multiple datasets
        excl_dict : dict, optional
            Dictionary of exclusion LayerMask arugments {layer: {kwarg: value}}
            by default None
        area_filter_kernel : str, optional
            Contiguous area filter method to use on final exclusions mask,
            by default "queen"
        min_area : float, optional
            Minimum required contiguous area filter in sq-km,
            by default None
        resolution : int, optional
            SC resolution, must be input in combination with gid. Prefered
            option is to use the row/col slices to define the SC point instead,
            by default 0.0081
        excl_area : float, optional
            Area of an exclusion pixel in km2. None will try to infer the area
            from the profile transform attribute in excl_fpath,
            by default None
        gids : list, optional
            List of gids to get summary for (can use to subset if running in
            parallel), or None for all gids in the SC extent, by default None
        gen_index : np.ndarray, optional
            Array of generation gids with array index equal to resource gid.
            Array value is -1 if the resource index was not used in the
            generation run, by default None
        kwargs : dict
            Unused kwargs from Aggregation.run_serial method, namely agg_method

        Returns
        -------
        agg_out : dict
            Aggregated values for each aggregation dataset
        """
        with SupplyCurveExtent(excl_fpath, resolution=resolution) as sc:
            exclusion_shape = sc.exclusions.shape
            if gids is None:
                gids = sc.valid_sc_points(tm_dset)
            elif np.issubdtype(type(gids), np.number):
                gids = [gids]

            slice_lookup = sc.get_slice_lookup(gids)

        cls._check_inclusion_mask(inclusion_mask, gids, exclusion_shape)

        # pre-extract handlers so they are not repeatedly initialized
        file_kwargs = {'excl_dict': excl_dict,
                       'area_filter_kernel': area_filter_kernel,
                       'min_area': min_area}
        dsets = wind_dir_dset + ('meta', )
        agg_out = {ds: [] for ds in dsets}
        with AggFileHandler(excl_fpath, h5_fpath, **file_kwargs) as fh:
            n_finished = 0
            for gid in gids:
                gid_inclusions = cls._get_gid_inclusion_mask(
                    inclusion_mask, gid, slice_lookup,
                    resolution=resolution)
                try:
                    gid_out = MeanWindDirectionsPoint.run(
                        gid,
                        fh.exclusions,
                        fh.h5,
                        tm_dset,
                        *wind_dir_dset,
                        excl_dict=excl_dict,
                        inclusion_mask=gid_inclusions,
                        resolution=resolution,
                        excl_area=excl_area,
                        exclusion_shape=exclusion_shape,
                        close=False,
                        gen_index=gen_index)

                except EmptySupplyCurvePointError:
                    logger.debug('SC gid {} is fully excluded or does not '
                                 'have any valid source data!'.format(gid))
                except Exception:
                    logger.exception('SC gid {} failed!'.format(gid))
                    raise
                else:
                    n_finished += 1
                    logger.debug('Serial aggregation: '
                                 '{} out of {} points complete'
                                 .format(n_finished, len(gids)))
                    log_mem(logger)
                    for k, v in gid_out.items():
                        agg_out[k].append(v)

        return agg_out

    def aggregate(self, max_workers=None, sites_per_worker=1000):
        """
        Average wind directions to sc_points

        Parameters
        ----------
        max_workers : int | None
            Number of cores to run summary on. None is all
            available cpus.
        sites_per_worker : int, optional
            Number of SC points to process on a single parallel worker,
            by default 1000

        Returns
        -------
        agg : dict
            Aggregated values for each aggregation dataset
        """
        agg = super().aggregate(max_workers=max_workers,
                                sites_per_worker=sites_per_worker)

        return agg

    @classmethod
    def run(cls, res_h5_fpath, excl_fpath, wdir_dsets,
            tm_dset='techmap_wtk', excl_dict=None,
            area_filter_kernel='queen', min_area=None,
            resolution=128, excl_area=None, max_workers=None,
            sites_per_worker=1000, out_fpath=None):
        """
        Aggregate powerrose to supply curve points, find neighboring supply
        curve point gids and rank them based on prominent powerrose direction

        Parameters
        ----------
        res_h5_fpath : str
            Filepath to .h5 file containing wind direction data
        excl_fpath : str
            Filepath to exclusions h5 with techmap dataset.
        wdir_dsets : str | list
            Wind direction dataset to average
        tm_dset : str, optional
            Dataset name in the techmap file containing the
            exclusions-to-resource mapping data,
            by default 'techmap_wtk'
        excl_dict : dict | None, optional
            Dictionary of exclusion LayerMask arugments {layer: {kwarg: value}}
            by default None
        area_filter_kernel : str, optional
            Contiguous area filter method to use on final exclusions mask,
            by default 'queen'
        min_area : float | None, optional
            Minimum required contiguous area filter in sq-km, by default None
        resolution : int | None, optional
            SC resolution, must be input in combination with gid,
            by default 128
        excl_area : float | None, optional
            Area of an exclusion pixel in km2. None will try to infer the area
            from the profile transform attribute in excl_fpath,
            by default None
        max_workers : int | None, optional
            Number of cores to run summary on. None is all
            available cpus, by default None
        sites_per_worker : int, optional
            Number of SC points to process on a single parallel worker,
            by default 1000
        out_fpath : str
            Path to .h5 file to save aggregated data too

        Returns
        -------
        agg : dict
            Aggregated values for each aggregation dataset
        """
        wdir = cls(res_h5_fpath, excl_fpath, wdir_dsets, tm_dset=tm_dset,
                   excl_dict=excl_dict, area_filter_kernel=area_filter_kernel,
                   min_area=min_area, resolution=resolution,
                   excl_area=excl_area)

        agg = wdir.aggregate(max_workers=max_workers,
                             sites_per_worker=sites_per_worker)

        if out_fpath is not None:
            wdir.save_agg_to_h5(out_fpath, agg)

        return agg



# -*- coding: utf-8 -*-
"""
Sub-packaged to compute the least cost path to tranmission
"""
from .cost_creator import XmissionCostCreator
from .least_cost_paths import LeastCostPaths
from .least_cost_xmission import LeastCostXmission



# -*- coding: utf-8 -*-
"""
Module to temporaly aggregate solar and wind CF profiles
"""
import logging
import numpy as np
import os
import pandas as pd

from reVX.handlers.outputs import Outputs
from rex.resource import Resource
from rex.utilities.execution import SpawnProcessPool
from rex.utilities.utilities import slice_sites, roll_timeseries

logger = logging.getLogger(__name__)


def get_time_index(h5_path, year=None):
    """
    Extract time_index from .h5 file. If 'year' is
    provided extact time_index for given year.

    Parameters
    ----------
    h5_path : str
        Path to .h5 file to extract meta and time_index
    year : str | int, optional
        Year to extract time-index for if running on a multi-year file,
        by default None

    Returns
    -------
    time_index: pandas.DatatimeIndex
        Datetime Index
    """
    with Resource(h5_path) as f:
        if 'time_index' in f:
            time_index = f.time_index
        elif year is not None:
            time_index = f[f'time_index-{year}']
        else:
            ti_dsets = [dset for dset in f.datasets
                        if dset.startswith('time_index')]
            msg = ("'time_index' is not available in {}. The following "
                   "potential annual time_index are available: {}. Please "
                   "specify a 'year' to use the 'time_index' for a "
                   "specific year.".format(h5_path, ti_dsets))
            logger.error(ti_dsets)
            raise ValueError(msg)

    return time_index


class DatasetAgg():
    """
    Temporaly Aggregate Dataset
    """
    def __init__(self, h5_fpath, dset, time_index=None, year=None,
                 local_time=False):
        """
        Parameters
        ----------
        h5_fpath : str
            Path to source h5 filepath
        dset : str
            Dataset to aggregate
        time_index : pandas.DatetimeIndex, optional
            Dataset datetime index, if None, extract from h5_fpath,
            by default None
        year : str | int, optional
            Year to extract time-index for if running on a multi-year file,
            by default None
        local_time : bool
            Flag to shift data to local time before aggregating temporal data.
            Default is to stay in UTC.
        """
        self._h5_fpath = h5_fpath
        self._dset = dset
        if time_index is None:
            time_index = get_time_index(h5_fpath, year=year)

        self._time_index = time_index
        self._year = year
        self._local_time = local_time

        with Resource(h5_fpath) as f:
            msg = '{} is not present in {}'.format(dset, h5_fpath)
            assert dset in f, msg

    def __repr__(self):
        msg = ('{} of {} in {}'
               .format(self.__class__.__name__, self._dset, self._h5_fpath))

        return msg

    @staticmethod
    def _aggregate_sites(h5_fpath, dset, site_slice=None, freq='1d',
                         method='mean', time_index=None, year=None,
                         local_time=False, **resample_kwargs):
        """
        Aggregate given sites in dataset

        Parameters
        ----------
        h5_fpath : str
            Path to source h5 filepath
        dset : str
            Dataset to aggregate
        site_slice : slice, optional
            Sites to aggregate, if None aggregate all sites, by default None
        freq : str, optional
            Aggregation frequency, by default '1d'
        method : str, optional
            Aggregation method, either 'mean' or 'sum', by default 'mean'
        time_index : pandas.DatetimeIndex, optional
            Dataset datetime index, if None, extract from h5_fpath,
            by default None
        year : str | int, optional
            Year to extract time-index for if running on a multi-year file,
            by default None
        local_time : bool
            Flag to shift data to local time before aggregating temporal data.
            Default is to stay in UTC.
        resample_kwargs : dict, optional
            Kwargs for pandas.DataFrame.resample

        Returns
        -------
        """
        if site_slice is None:
            site_slice = slice(None)

        if time_index is None:
            time_index = get_time_index(h5_fpath, year=year)

        with Resource(h5_fpath) as f:
            arr = f[dset, :, site_slice]

            if local_time:
                timezones = f.get_meta_arr('timezone', rows=site_slice)
                arr = roll_timeseries(arr, timezones)

        agg_data = pd.DataFrame(arr, index=time_index)
        agg_data = agg_data.resample(freq, **resample_kwargs)

        if method.lower() == 'mean':
            agg_data = agg_data.mean()
        elif method.lower() == 'sum':
            agg_data = agg_data.sum()
        else:
            msg = ('Invalid aggregation method {}, must be "mean" or '
                   '"sum"!'.format(method))
            logger.error(msg)
            raise ValueError(msg)

        return agg_data.values

    def _get_slices(self, dset, chunks_per_slice=5):
        """
        Get slices to extract

        Parameters
        ----------
        dset : str
            Dataset to aggregate
        chunks_per_slice : int, optional
            Number of chunks to extract in each slice, by default 5

        Returns
        -------
        slices : list
            List of slices to extract
        """
        with Resource(self._h5_fpath) as f:
            shape, _, chunks = f.get_dset_properties(dset)

        if len(shape) != 2:
            msg = ('Cannot aggregate dataset {}, as it is '
                   'not a timeseries dataset!'.format(dset))
            logger.error(msg)
            raise RuntimeError(msg)

        slices = slice_sites(shape, chunks,
                             chunks_per_slice=chunks_per_slice)

        return slices, shape

    def _create_agg_arr(self, shape, freq='1d', **resample_kwargs):
        """
        Create empty aggregation array

        Parameters
        ----------
        shape : tuple
            Dataset shape
        freq : str, optional
            Aggregation frequency, by default '1d'
        resample_kwargs : dict, optional
            Kwargs for pandas.DataFrame.resample

        Returns
        -------
        out : ndarray
            Output array for aggregated dataset
        """
        out = pd.Series(0, index=self._time_index)
        out = out.resample(freq, **resample_kwargs).mean()
        shape = (out.shape[0], shape[1])

        return np.zeros(shape, dtype=np.float32)

    def aggregate(self, freq='1d', method='mean', max_workers=None,
                  chunks_per_worker=5, **resample_kwargs):
        """
        Aggregate dataset to desired frequency using desired method

        Parameters
        ----------
        freq : str, optional
            Aggregation frequency, by default '1d'
        method : str, optional
            Aggregation method, either 'mean' or 'sum', by default 'mean'
        max_workers : None | int, optional
            Number of workers to use, if 1 run in serial, if None use all
            available cores, by default None
        chunks_per_worker : int, optional
            Number of chunks to extract on each worker, by default 5
        resample_kwargs : dict, optional
            Kwargs for pandas.DataFrame.resample

        Returns
        -------
        dset_agg : ndarray
            Aggregated dataset array
        """
        if max_workers is None:
            max_workers = os.cpu_count()

        slices, shape = self._get_slices(self._dset,
                                         chunks_per_slice=chunks_per_worker)
        if len(slices) == 1:
            max_workers = 1

        dset_agg = self._create_agg_arr(shape, freq=freq, **resample_kwargs)

        if max_workers > 1:
            msg = ('Aggregating {} in parallel using {} workers'
                   .format(self._dset, max_workers))
            logger.info(msg)

            loggers = [__name__, 'rex', 'reVX']
            with SpawnProcessPool(max_workers=max_workers,
                                  loggers=loggers) as exe:
                futures = []
                for sites_slice in slices:
                    future = exe.submit(self._aggregate_sites,
                                        self._h5_fpath,
                                        self._dset,
                                        site_slice=sites_slice,
                                        freq=freq,
                                        method=method,
                                        time_index=self._time_index,
                                        year=self._year,
                                        local_time=self._local_time,
                                        **resample_kwargs)
                    futures.append(future)

                for i, future in enumerate(futures):
                    dset_agg[:, slices[i]] = future.result()
                    logger.debug('Completed {} out of {} workers'
                                 .format((i + 1), len(futures)))
        else:
            msg = 'Aggregating {} in serial'.format(self._dset)
            logger.info(msg)
            for i, sites_slice in enumerate(slices):
                dset_agg[:, sites_slice] = self._aggregate_sites(
                    self._h5_fpath,
                    self._dset,
                    site_slice=sites_slice,
                    freq=freq,
                    method=method,
                    time_index=self._time_index,
                    year=self._year,
                    local_time=self._local_time,
                    **resample_kwargs)
                logger.debug('Completed {} out of {} sets of sites'
                             .format((i + 1), len(slices)))

        return dset_agg

    @classmethod
    def run(cls, h5_fpath, dset, time_index=None, year=None, freq='1d',
            method='mean', max_workers=None, chunks_per_worker=5,
            local_time=False, **resample_kwargs):
        """
        Temporally aggregate dataset to given frequency using given method

        Parameters
        ----------
        h5_fpath : str
            Path to source h5 filepath
        dset : str
            Dataset to aggregate
        time_index : pandas.DatetimeIndex, optional
            Dataset datetime index, if None, extract from h5_fpath,
            by default None
        year : str | int, optional
            Year to extract time-index for if running on a multi-year file,
            by default None
        freq : str, optional
            Aggregation frequency, by default '1d'
        method : str, optional
            Aggregation method, either 'mean' or 'sum', by default 'mean'
        max_workers : None | int, optional
            Number of workers to use, if 1 run in serial, if None use all
            available cores, by default None
        chunks_per_worker : int, optional
            Number of chunks to extract on each worker, by default 5
        local_time : bool
            Flag to shift data to local time before aggregating temporal data.
            Default is to stay in UTC.
        resample_kwargs : dict, optional
            Kwargs for pandas.DataFrame.resample

        Returns
        -------
        agg_data : ndarray
            Dataset aggregated do given frequency using given method
        """
        logger.info('Aggregating {} in {} to {} using {}'
                    .format(dset, h5_fpath, freq, method))
        agg = cls(h5_fpath, dset, time_index=time_index, year=year,
                  local_time=local_time)
        agg_dset = agg.aggregate(freq=freq, method=method,
                                 max_workers=max_workers,
                                 chunks_per_worker=chunks_per_worker,
                                 **resample_kwargs)

        return agg_dset


class TemporalAgg():
    """
    Class to temporally aggregate time-series data
    """
    def __init__(self, src_fpath, dst_fpath, freq='1d', dsets=None, year=None,
                 local_time=False, **resample_kwargs):
        """
        Parameters
        ----------
        src_fpath : str
            Path to source h5 file
        dst_fpath : str
            Path to destination h5 file to save aggregated datasets to.
        freq : str, optional
            Aggregation frequency, by default '1d'
        dsets : list, optional
            Datasets to aggregate, if None aggregate all datasets in src_fpath,
            by default None
        year : str | int, optional
            Year to extract time-index and datasets for, needed  if running
            on a multi-year file, by default None
        local_time : bool
            Flag to shift data to local time before aggregating temporal data.
            Default is to stay in UTC.
        resample_kwargs : dict, optional
            Kwargs for pandas.DataFrame.resample
        """
        self._src_fpath = src_fpath
        self._dst_fpath = dst_fpath
        self._freq = freq
        self._local_time = local_time
        self._dsets, self._time_index = self._get_dsets(src_fpath, dsets=dsets,
                                                        year=year)
        self._resample_kwargs = resample_kwargs

        self._init_agg_h5(year=year)

    def __repr__(self):
        msg = '{} of {}'.format(self.__class__.__name__, self._src_fpath)

        return msg

    @property
    def dsets(self):
        """
        Datasets to aggregate

        Returns
        -------
        list
        """
        return self._dsets

    @staticmethod
    def _get_dsets(h5_fpath, dsets=None, year=None):
        """
        Get datasets to aggregate, or if given ensure they are in source
        h5 file. Also extract time_index

        Parameters
        ----------
        h5_fpath : str
            Path to .h5 file to source dataset for aggregation from
        dsets : list, optional
            Datasets to aggregate, if None aggregate all datasets in src_fpath,
            by default None


        Returns
        -------
        dsets : list
            List of datasets to aggregate
        time_index : pandas.DatetimeIndex
            DatetimeIndex of datasets being aggregated
        """
        with Resource(h5_fpath) as f:
            if dsets is None:
                res_dsets = f.resource_datasets
                if year is not None:
                    res_dsets = [ds for ds in res_dsets
                                 if str(year) in ds
                                 and 'time_index' not in ds]

                dsets = []
                for ds in res_dsets:
                    if len(f.shapes[ds]) == 2:
                        dsets.append(ds)

            else:
                for ds in dsets:
                    msg = "{} is not available in {}".format(ds, h5_fpath)
                    assert ds in f, msg

        time_index = get_time_index(h5_fpath, year=year)

        return dsets, time_index

    def _init_agg_h5(self, year=None):
        """
        Initialize the dst .h5 file that will contain the aggregated datasets

        Parameter
        ---------
        year : str | int, optional
            Year to extract time-index and datasets for, needed  if running
            on a multi-year file, by default None
        """
        logger.info('Initializing {}'.format(self._dst_fpath))
        time_index = pd.Series(0, index=self._time_index)
        time_index = time_index.resample(self._freq, **self._resample_kwargs)
        time_index = time_index.mean().index
        shape = (len(time_index), )
        ti_dset = 'time_index'

        if year is not None:
            ti_dset += '-{}'.format(year)

        with Outputs(self._dst_fpath, mode='a') as f_out:
            with Resource(self._src_fpath) as f_in:
                meta = f_in.meta
                shape += (len(meta), )
                f_out._shape = shape
                if 'meta' not in f_out:
                    logger.debug('Copying meta data')
                    f_out['meta'] = meta

                logger.debug('Copying {}'.format(ti_dset))
                f_out._set_time_index(ti_dset, time_index)

                for ds in self.dsets:
                    _, dtype, chunks = f_in.get_dset_properties(ds)
                    attrs = f_in.attrs[ds]
                    logger.debug('Initializing aggregated {} w/ properties:'
                                 '\nshape: {}'
                                 '\ndtype: {}'
                                 '\nchunks: {}'
                                 '\nattrs: {}'
                                 .format(ds, shape, dtype, chunks, attrs))

                    f_out._create_dset(ds, shape, dtype, chunks=chunks,
                                       attrs=attrs)

    def aggregate(self, method='mean', max_workers=None, chunks_per_worker=5):
        """
        Aggregate desired datasets and write to disk

        Parameters
        ----------
        method : str, optional
            Aggregation method, either 'mean' or 'sum', by default 'mean'
        max_workers : None | int, optional
            Number of workers to use, if 1 run in serial, if None use all
            available cores, by default None
        chunks_per_worker : int, optional
            Number of chunks to extract on each worker, by default 5
        """
        for ds in self.dsets:
            ds_agg = DatasetAgg.run(self._src_fpath, ds,
                                    time_index=self._time_index,
                                    freq=self._freq,
                                    method=method,
                                    max_workers=max_workers,
                                    chunks_per_worker=chunks_per_worker,
                                    local_time=self._local_time,
                                    **self._resample_kwargs)
            with Outputs(self._dst_fpath, mode='a') as f:
                logger.info('Writing aggregated data for {} to disk'
                            .format(ds))
                f[ds] = ds_agg

    @classmethod
    def run(cls, src_fpath, dst_fpath, freq='1d', dsets=None, year=None,
            method='mean', max_workers=None, chunks_per_worker=5,
            local_time=False, **resample_kwargs):
        """
        Temporally aggregate the desired datasets in the src .h5 file to the
        given frequency using the given method. Save the aggregated datasets
        to the dst .h5 file.

        Parameters
        ----------
        src_fpath : str
            Path to source h5 file
        dst_fpath : str
            Path to destination h5 file to save aggregated datasets to.
        freq : str, optional
            Aggregation frequency, by default '1d'
        dsets : list, optional
            Datasets to aggregate, if None aggregate all datasets in src_fpath,
            by default None
        year : str | int, optional
            Year to extract time-index and datasets for, needed  if running
            on a multi-year file, by default None
        method : str, optional
            Aggregation method, either 'mean' or 'sum', by default 'mean'
        max_workers : None | int, optional
            Number of workers to use, if 1 run in serial, if None use all
            available cores, by default None
        chunks_per_worker : int, optional
            Number of chunks to extract on each worker, by default 5
        local_time : bool
            Flag to shift data to local time before aggregating temporal data.
            Default is to stay in UTC.
        resample_kwargs : dict, optional
            Kwargs for pandas.DataFrame.resample
        """
        logger.info('Aggregating datasets in {} to {} using {} and saving to '
                    '{}'.format(src_fpath, freq, method, dst_fpath))
        agg = cls(src_fpath, dst_fpath, freq=freq, dsets=dsets, year=year,
                  local_time=local_time, **resample_kwargs)
        agg.aggregate(method=method, max_workers=max_workers,
                      chunks_per_worker=chunks_per_worker)




# -*- coding: utf-8 -*-
"""
Compute setbacks
"""
from .wind_setbacks import (StructureWindSetbacks, RoadWindSetbacks,
                            RailWindSetbacks, TransmissionWindSetbacks)
from .solar_setbacks import ParcelSetbacks
